#!/usr/bin/env python3
"""
indexff
objective: generalized positional indexer for entries in a flatfile

by Giuseppe Insana, 2020
"""

# pylint: disable=C0103,R0912,R0915,W0603

import os
import sys
import re
import time
import getpass
import argparse
from heapq import merge #mergesort for merging at the end of multithread
from multiprocessing import Pool, Array #for multithreaded
from random import randint #for compressed flatfiles via gztool
from tqdm import tqdm #for progress bar
from sortedcontainers import SortedList #if sorted output
from ffdb import eprint, derive_key, \
    get_cipher_type, format_indexes, check_iofiles, int_to_b64, \
    calculate_blocksize, siprefix2num, calculate_chunknum, \
    print_subfiles, delete_files, compute_split_positions, elapsed_time, \
    entry_generator, shift_index_file, buffered_readwrite, TEMPDIR, \
    postprocess_entry, PROGRESSBARCHARS

#CUSTOMIZATIONS:
#you can modify the default size of block (chunk) of flatfile
#to work on for parallel execution
MAXBLOCKSIZE = "50M" #50 Mb of flatfile in each chunk
MINBLOCKSIZE = "100k" #100k minimum each chunk

#CONSTANTS
PROGNAME = "indexer.py"
VERSION = "4.0"
AUTHOR = "Giuseppe Insana"
args = None
patterns = None
joinedpatterns = None


def check_args():
    """
    parse arguments and check for error conditions
    """
    global args, patterns, joinedpatterns
    usagetxt = """{0} -f FLATFILE -i 'PATTERN' [-e ENDPATTERN] >INDEXFILE
    [-f] : flatfile to index
    [-i] : regex pattern for the identifiers; also [-j], see examples below
    [-e] : pattern for end of entry. defaults to "^-$"
    see '{0} -h' for tweaks and optional modes
    \nnotes: If compression or encryption is requested, an output flatfile will
         be created, and the resulting index will refer to it.
       If the identifiers are a LOT and memory is an issue, you may wish to use
         [-u] option and sort the resulting index after it has been generated.\nexamples:
       {0} -i '^AC   (.+?);' -f uniprot.dat -e '^//$' >up.pac
       {0} -i '^AC   (.+?);' 'ID   (.+?);' -f [...]
         (multiple patterns can be specified)

       {0} -i '^AC   (.+?);' -j '^OX   NCBI_(Tax)ID=(\\d+) ' -f [...]
         (complex patterns made of multiple parts can be specified with [-j];
          -i and -j patterns can be used together)

       {0} -a -j '^DR   (.+?);( .+?);' -f [...]
       {0} -a -i '^AC   (\\S+?); ?(\\S+)?;? ?(\\S+)?;?' -f [...]
         (use [-a] option to find all instances and capture groups of
          the provided patterns, not just the first one)
    """.format(PROGNAME)
    parser = argparse.ArgumentParser(description='Create a positional index for any flatfile, \
                                     optionally compressing or encrypting its entries',
                                     usage=usagetxt)
    parser.add_argument('-f', '--file', dest='input_filename',
                        help="Filename of flatfile to be processed", required=True)
    parser.add_argument('-i', '--id', dest='patterns',
                        help="regexp pattern for identifier(s) to index", required=False,
                        type=str, nargs='+')
    parser.add_argument('-j', '--joinedid', dest='joinedpatterns',
                        help="regexp pattern for identifier(s) to index", required=False,
                        type=str, nargs='+')
    parser.add_argument('-e', '--endpattern', dest='terminator',
                        help="regexp pattern to identify the end of each entry. If unspecified \
                        it defaults to '^-$'", required=False)
    parser.add_argument('-a', '--allmatches', dest='allmatches', action='store_true',
                        help="find all instances of the identifier pattern, not just the first one \
                        (the default behaviour)", required=False)
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help="verbose operation", required=False)
    parser.add_argument('-t', '--threads', dest='threads',
                        help="use specified number of threads for parallel indexing",
                        required=False, type=int)
    parser.add_argument('-b', '--blocksize', dest='input_blocksize',
                        help="redefine blocksize used for parallel execution. By default \
                        it will be adjusted automatically to the number of threads",
                        required=False, type=siprefix2num)
    parser.add_argument('-o', '--offset', dest='pos_offset',
                        help="optional offset (in bytes) to shift entry positions in index",
                        required=False, type=int)
    parser.add_argument('-k', '--keysize', dest='keysize',
                        help="request entries to be encrypted and specify encryption strength: \
                        16=aes-128, 24=aes-192 or 32=aes-256. INPUT_FILENAME.enc will be created",
                        required=False, type=int, choices=(16, 24, 32))
    parser.add_argument('-p', '--passphrase', dest='passphrase',
                        help="passphrase for encrypting the entries; if unspecified it will be \
                        requested interactively (safer)", required=False, type=str)
    parser.add_argument('-c', '--compresslevel', dest='compresslevel',
                        help="request entries to be compressed and specify a compress level. \
                        INPUT_FILENAME.xz will be created",
                        required=False, type=int, choices=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9))
    parser.add_argument('-x', '--xsanity', dest='xsanity', action='store_true',
                        help="compute entry checksums and add them to index",
                        required=False)
    parser.add_argument('-u', '--unsorted', dest='unsorted', action='store_true',
                        help="do not sort the index, leaving that task to a followup external \
                        command. Note that extraction requires a sorted index",
                        required=False)
    parser.add_argument('-n', '--nopos', dest='nopos', action='store_true',
                        help="do not compute positions, just print matching identifiers",
                        required=False)
    args = parser.parse_args()

    randnum = str(randint(1000, 9999))

    if args.patterns is None and args.joinedpatterns is None:
        eprint("    => ERROR: at least one of -i or -j needs to be given!")
        sys.exit(22)

    if args.patterns is None:
        args.patterns = []

    if args.joinedpatterns is None:
        args.joinedpatterns = []

    args.progressbar = False
    if args.verbose:
        eprint(" .-- {} v{} -- by {} --.".format(PROGNAME, VERSION, AUTHOR))
        args.progressbar = True

    if args.allmatches:
        if args.verbose:
            eprint(" |-- All matches of the pattern will be stored as identifiers")

    if args.nopos:
        if args.pos_offset is not None or args.compresslevel is not None \
                or args.passphrase is not None or args.keysize is not None or \
                args.xsanity:
            eprint("    => ERROR: No sense specifying compression, encryption, sanity")
            eprint("    =>        or pos_offset when using --nopos option")
            sys.exit(22)
        if args.verbose:
            eprint(" |-- No positional index will be created. Only printing identifiers found")

    if args.pos_offset is None:
        args.pos_offset = 0
    else:
        if args.verbose:
            eprint(" |-- positions to be offset by: {}".format(args.pos_offset))

    if args.unsorted:
        if args.verbose:
            eprint("    => NOTICE: index will be printed unsorted as requested.")
            eprint("       Please sort index before using it for extraction")

    patterns = set()
    for pattern in args.patterns:
        if args.verbose:
            eprint(" |-- adding identifier pattern '{}'".format(pattern))
        patterns.add(re.compile(pattern.encode('UTF-8'), re.MULTILINE))
    joinedpatterns = set()
    for pattern in args.joinedpatterns:
        if args.verbose:
            eprint(" |-- adding joined identifier pattern '{}'".format(pattern))
        joinedpatterns.add(re.compile(pattern.encode('UTF-8'), re.MULTILINE))

    if args.terminator is None:
        args.terminator = "^-$" #default
    if args.verbose:
        eprint(" |-- entry terminator pattern set as '{}'".format(args.terminator))

    if args.xsanity and args.verbose:
        eprint(" |-- entry checksums will be computed and added to index")

    if args.keysize is not None:
        if args.passphrase is None:
            eprint(" |-- keysize specified, please provide a passphrase:")
            args.passphrase = getpass.getpass(prompt=" |>> Passphrase: ")

    args.encrypt = False
    args.cipher_type = None
    args.key = None
    if args.passphrase is not None:
        args.encrypt = True
        if args.keysize is None:
            args.keysize = 16 #default
        args.cipher_name, args.key = derive_key(args.passphrase, args.keysize)
        args.cipher_type = get_cipher_type(args.cipher_name)
        if args.verbose:
            eprint(" |-- encrypted flatfile (with cipher {}) will be written to '{}.enc'".format(
                args.cipher_name, args.input_filename))
            #eprint(" |-- the passphrase is: {}".format(args.passphrase)) #DEBUG!
            #eprint(" |-- the encryption key is: {}".format(args.key)) #DEBUG!

    args.compress = False
    if args.compresslevel is not None:
        args.compress = True
        if args.encrypt:
            args.index_type = "+" #both compressed and encrypted
            if args.verbose:
                eprint(" |-- entries will be compressed and encrypted to .enc file")
        else:
            args.index_type = ":" #compressed but not encrypted
            if args.verbose:
                eprint(" |-- entries will be compressed to .xz file")
    else: #not compressed
        if args.encrypt:
            args.index_type = "." #encrypted but not compressed
            if args.verbose:
                eprint(" |-- entries will be encrypted to .enc file")
        else:
            args.index_type = "-" #not encrypted nor compressed but entry sizes stored

    if args.input_blocksize is not None:
        if args.threads is None:
            eprint("    => ERROR: specifying blocksize makes sense only for -t execution")
            sys.exit(22)
        if args.verbose:
            eprint(" |-- blocksize set to {} bytes".format(args.input_blocksize))
    if args.threads is not None: #multithread
        if args.threads < 2:
            eprint("    => ERROR: No sense specifying a number of threads lower than 2!")
            sys.exit(22)
        if args.input_blocksize is None:
            #if not specified, we use 1/threadnumTH of inputfilesize up
            #to a maximum MAXBLOCKSIZE and with minimum MINBLOCKSIZE
            args.input_blocksize = max(siprefix2num(MINBLOCKSIZE),
                                       min(
                                           calculate_blocksize(args.input_filename, args.threads),
                                           siprefix2num(MAXBLOCKSIZE)))
        args.input_filesize = os.path.getsize(args.input_filename)
        if args.input_blocksize > args.input_filesize // 2 * 3:
            eprint("    => NOTICE: blocksize too BIG compared to flatfile size, -t not applicable!")
            sys.exit(22)
        args.mt_subfiles_dir = TEMPDIR + "/tmpINDEX" + randnum + "/"
        args.mt_subfiles_fprefix = args.mt_subfiles_dir + "F"
        args.mt_subfiles_iprefix = args.mt_subfiles_dir + "I"
        args.mt_subfiles_oprefix = args.mt_subfiles_dir + "O"
        #find max number of chunks required (we'll adjust later on split)
        args.input_filesize, args.chunks_count = calculate_chunknum(args.input_filename,
                                                                    args.input_blocksize)
        if args.verbose:
            eprint(" |-- using maximum {} parallel threads (-t); your OS reports {} cpus.".format(
                args.threads, os.cpu_count()))
    else: # if unspecified, set args.threads to 1
        args.threads = 1
        args.chunks_count = 1

    if args.nopos:
        args.index_type = "" #no indexes


def check_files():
    """
    check for ability to open input/output filenames
    """
    check_iofiles([args.input_filename], [])

    args.outff_filename = None
    if args.encrypt:
        args.outff_filename = "{}.{}".format(args.input_filename, "enc")
    elif args.compress: #if compressed but not encrypted
        args.outff_filename = "{}.{}".format(args.input_filename, "xz")

    if args.outff_filename is not None:
        check_iofiles([], [args.outff_filename])

    if args.threads > 1: #multithread
        os.mkdir(args.mt_subfiles_dir, mode=0o700)
        args.chunk_itempfiles = list()
        for chunknum in range(args.chunks_count):
            chunknumstr = str(chunknum).zfill(len(str(args.chunks_count))) #e.g. 00, 01..
            chunk_tempfile = args.mt_subfiles_iprefix + "." + chunknumstr
            try:
                myoutputfh = open(chunk_tempfile, 'w')
            except PermissionError:
                eprint("    => ERROR: Cannot open temporary file '{}' for writing".format(
                    chunk_tempfile))
                sys.exit(1)
            args.chunk_itempfiles.append(chunk_tempfile) #store temp filenames
            myoutputfh.close()
        delete_files(args.chunk_itempfiles)
        #for chunknum in range(args.chunks_count): #DEBUG!
        #   eprint(" >> the temporary index file for chunk #{} will be '{}'".format(
        #       chunknum, args.chunk_itempfiles[chunknum]))

        #if outff needs to be generated
        if args.compress or args.encrypt:
            args.chunk_otempfiles = list()
            for chunknum in range(args.chunks_count):
                chunknumstr = str(chunknum).zfill(len(str(args.chunks_count))) #e.g. 00, 01..
                chunk_tempfile = args.mt_subfiles_oprefix + "." + chunknumstr
                try:
                    myoutputfh = open(chunk_tempfile, 'w')
                except PermissionError:
                    eprint("    => ERROR: Cannot open temporary file '{}' for writing".format(
                        chunk_tempfile))
                    sys.exit(1)
                args.chunk_otempfiles.append(chunk_tempfile) #store temp filenames
                myoutputfh.close()
            delete_files(args.chunk_otempfiles)


def find_patterns_in_entry(entry, allmatches=False):
    """
    parse an entry, using regular expressions to capture desired fields
    """
    goodidentifiers = list()
    if allmatches:
        for pattern in patterns:
            identifiers = pattern.findall(entry['full'])
            #eprint("trying {} found {}".format(pattern, identifiers))
            for match in identifiers:
                if isinstance(match, tuple):
                    for submatch in match: #uglier but faster than list comprehension
                        if submatch:
                            goodidentifiers.append(submatch.decode())
                else:
                    goodidentifiers.append(match.decode())
        for pattern in joinedpatterns:
            identifiers = pattern.findall(entry['full'])
            #eprint("trying {} found {}".format(pattern, identifiers))
            for match in identifiers:
                if isinstance(match, tuple):
                    goodidentifiers.append(b''.join(match).decode())
                else:
                    goodidentifiers.append(match.decode())
    else:
        for pattern in patterns:
            match = pattern.search(entry['full'])
            if match:
                for submatch in match.groups():
                    if submatch:
                        goodidentifiers.append(submatch.decode())
        for pattern in joinedpatterns:
            match = pattern.search(entry['full'])
            if match:
                submatches = b''
                for submatch in match.groups():
                    if submatch:
                        submatches += submatch
                goodidentifiers.append(submatches.decode())

    entry['ids'].update(goodidentifiers)


def parse_ff_wrapper(chunknum):
    """
    wrapper to parse_ff for multithread use
    """
    #collect all index parameters in mode dict
    mode = {'encrypt': args.encrypt, 'compress': args.compress,
            'key': args.key, 'index_type': args.index_type,
            'cipher_type': args.cipher_type,
            'compresslevel': args.compresslevel,
            'terminator': args.terminator,
            'unsorted': args.unsorted,
            'xsanity': args.xsanity,
            'allmatches': args.allmatches,
            'nopos': args.nopos
            }

    #multithreaded will write each chunk index to a temporary file
    outindex_filename = None
    if args.threads > 1:
        outindex_filename = args.chunk_itempfiles[chunknum]

    #if encryption or compression, outfile will be generated
    outff_filename = None
    if args.threads > 1:
        if args.encrypt or args.compress:
            outff_filename = args.chunk_otempfiles[chunknum]
    else:
        outff_filename = args.outff_filename

    #compute pos_offset
    pos_offset = 0
    if args.threads > 1:
        input_file = args.chunk_ftemp_files[chunknum] #split chunk of flatfile
        buffered_readwrite(args.input_filename,
                           input_file,
                           args.chunk_ftemp_startpos[chunknum],
                           args.chunk_ftemp_filesizes[chunknum])

        pos_offset = args.pos_offset
        if chunknum > 0: #no need for first chunk
            if not args.compress: #if compress we'll shift indexes later
                for prev_chunknum in range(chunknum):
                    pos_offset += args.chunk_ftemp_filesizes[prev_chunknum]
            #eprint("#{}: adding prev chunks filesize, offset now {}".format( #DEBUG
            #   chunknum, pos_offset))
    else:
        input_file = args.input_filename
        pos_offset = args.pos_offset
    mode['pos_offset'] = pos_offset

    parse_ff_byentry(input_file,
                     mode,
                     outindex_filename=outindex_filename,
                     outff_filename=outff_filename,
                     chunknum=chunknum)


def parse_ff_byentry(inputfile, mode, outff_filename=None,
                     outindex_filename=None, chunknum=0):
    """
    parse input file entry by entry
    """
    global entries_count, indexes_count, skipped_count
    if not mode['unsorted']:
        mode['indexes'] = SortedList()
    #init entry hash
    entry = dict()
    entry['ids'] = set()
    entry['full'] = "" #used for calculating entrysizes, compressing or encrypting
    entry['length'] = 0
    if outff_filename is not None:
        mode['outffh'] = open(outff_filename, 'wb')
    if outindex_filename is None:
        mode['outindexfh'] = sys.stdout #by default print entries to stdout
    else:
        mode['outindexfh'] = open(outindex_filename, 'w')
    entryposition = 0 #position of first entry
    if mode['pos_offset'] > 0:
        entryposition = mode['pos_offset']
        #eprint("applying pos_offset of {}".format(pos_offset)) #DEBUG
    bufsize = 1024 * 1024
    for entry['full'], entry['length'] in entry_generator(inputfile, mode['terminator'], bufsize):
        entries_count[chunknum] += 1
        find_patterns_in_entry(entry, mode['allmatches'])
        if entry['ids']: #if identifiers found
            if not mode['nopos']:
                #position and postprocessing
                entry['position'] = int_to_b64(entryposition)
                postprocess_entry(entry, mode)

            #index formatting
            new_indexes = format_indexes(entry, mode['index_type'],
                                         mode['cipher_type'], mode['xsanity'])
            if mode['unsorted']: #write out continously the indexes
                mode['outindexfh'].write("".join(new_indexes))
                indexes_count[chunknum] += len(new_indexes)
            else: #if sorted, update sortedlist, write indexes in the end
                mode['indexes'].update(new_indexes)
            entry['ids'] = set() #reset entry hash
        else: #we skip the entry since we found no identifiers
            skipped_count[chunknum] += 1
        if not mode['nopos']:
            entryposition += entry['length'] # for next entry

    if not mode['unsorted']: #if sorted, we write all now, at the end
        indexes_count[chunknum] = len(mode['indexes'])
        for index in mode['indexes']:
            mode['outindexfh'].write(index)
    if outff_filename is not None:
        mode['outffh'].close()
    if outindex_filename is not None:
        mode['outindexfh'].close()


def print_stats(start_time):
    """
    if verbose print some final statistics on entries indexed
    """
    entries_sum = 0
    indexes_sum = 0
    skipped_sum = 0
    for chunknum in range(args.chunks_count):
        entries_sum += entries_count[chunknum]
        indexes_sum += indexes_count[chunknum]
        skipped_sum += skipped_count[chunknum]
    eprint(" '-- {} entries with {} indexes ~ {} entries skipped --'".format(
        entries_sum, indexes_sum, skipped_sum))
    eprint(" '-- Elapsed: {}, {} entries/sec --'".format(*elapsed_time(start_time, entries_sum)))


def init_thread(e, i, s):
    """
    to initialise multithreaded worker
    """
    global entries_count, indexes_count, skipped_count
    entries_count = e
    indexes_count = i
    skipped_count = s


if __name__ == '__main__':
    check_args()
    check_files()

    start_secs = time.time()
    if args.threads > 1: #multithread
        args.chunk_ftemp_files = list()

        #find out where to split the input file without breaking entries
        args.chunk_ftemp_startpos, args.chunk_ftemp_filesizes = compute_split_positions(
            args.input_filename, args.input_blocksize, args.terminator)
        args.chunks_count = len(args.chunk_ftemp_filesizes)
        suffixlength = len(str(args.chunks_count))
        for mychunknum in range(args.chunks_count):
            chunk_suffix = str(mychunknum).zfill(suffixlength)
            args.chunk_ftemp_files.append(args.mt_subfiles_dir + chunk_suffix)

        if args.verbose:
            eprint(" |-- parallel work in chunks of maximum {} bytes (-b to change)".format(
                args.input_blocksize))
            eprint(" |-- flatfile will be split into {} chunks".format(args.chunks_count))

        entries_count = Array('i', args.chunks_count)
        indexes_count = Array('i', args.chunks_count)
        skipped_count = Array('i', args.chunks_count)

        args.chunk_itempfiles = args.chunk_itempfiles[0:args.chunks_count]
        if args.outff_filename is not None:
            args.chunk_otempfiles = args.chunk_otempfiles[0:args.chunks_count]

        #init threads
        pool = Pool(args.threads, initializer=init_thread, initargs=(
            entries_count, indexes_count, skipped_count))
        #submit chunks to threads
        if args.progressbar:
            _ = list(tqdm(pool.imap(parse_ff_wrapper, range(args.chunks_count)),
                          total=args.chunks_count, ascii=PROGRESSBARCHARS))
        else:
            pool.imap(parse_ff_wrapper, range(args.chunks_count))
        pool.close() #no more work to submit
        pool.join() #wait workers to finish

        delete_files(args.chunk_ftemp_files) #cleanup

        #final union of results:
        if args.compress:
            otempfile_offset = 0
            otempfile_offsets = list()
            # 1) go through .enc or .xz files and sum filesizes as offsets
            for otempfile in args.chunk_otempfiles:
                otempfile_offsets.append(otempfile_offset)
                otempfile_offset += os.path.getsize(otempfile)

            # 2) go through indexes (a part first one) and shift their offset
            for mychunknum in range(args.chunks_count):
                itempfile = args.chunk_itempfiles[mychunknum]
                outfile = itempfile + "s"
                shift_index_file(itempfile,
                                 otempfile_offsets[mychunknum], outfile,
                                 args.index_type)
                args.chunk_itempfiles[mychunknum] = outfile #replace
                os.remove(itempfile)

        #print merged index
        if args.unsorted: #simple printout
            print_subfiles(args.chunk_itempfiles)
        else: #merge sort and printout
            indexfilehandles = list()
            for filename in args.chunk_itempfiles:
                ifh = open(filename, 'r', 1)
                indexfilehandles.append(ifh)
            sys.stdout.writelines(merge(*indexfilehandles)) #heapq mergesort
        delete_files(args.chunk_itempfiles) #cleanup

        #concatenate final encrypted or compressed filefile, if requested
        if args.outff_filename is not None:
            with open(args.outff_filename, 'wb') as outputfh:
                print_subfiles(args.chunk_otempfiles, outputfh)
            delete_files(args.chunk_otempfiles) #cleanup

        if os.path.exists(args.mt_subfiles_dir):
            os.rmdir(args.mt_subfiles_dir) #cleanup
    else: #singlethread
        entries_count = Array('i', 1)
        indexes_count = Array('i', 1)
        skipped_count = Array('i', 1)

        parse_ff_wrapper(0)

    #final printout
    print_stats(start_secs)



#!/usr/bin/env python3
"""
extractor
objective: generalized entry extractor from local/remote/compressed/encrypted flatfile

by Giuseppe Insana, 2020
"""

# pylint: disable=C0103,R0912,R0915,W0603

#IMPORTS:
import os
import re
import sys
import time
import zlib
import getpass
import argparse
from math import ceil
from random import randint
from sortedcontainers import SortedList #if sorted output
#OPTIONAL IMPORTS (uncomment any of the following if never needed):
from subprocess import run, Popen, PIPE, DEVNULL #for compressed flatfiles via gztool
from multiprocessing import Pool, Array, current_process #for multithreaded
from tqdm import tqdm #for progress bar
from requests import get #for remote flatfiles via http
#NECESSARY IMPORTS
from ffdb import eprint, inflate, derive_key, check_index, \
    delete_files, REES, REESIV, GZTOOL_EXE, TEMPDIR, \
    init_cipher, check_iofiles, b64_to_int, \
    calculate_chunknum, calculate_blocksize, chunk_of_lines, siprefix2num, \
    read_from_size, print_subfiles, close_subfiles, elapsed_time, int_to_b64, \
    get_position_first, get_position_last, \
    get_position_checksum_first, get_position_checksum_last, \
    get_positions, get_positions_checksums, PROGRESSBARCHARS

#CUSTOMIZATIONS:
#you can modify the minimum default size of block (chunk) of identifiers list
#to work on for parallel execution
MINBLOCKSIZE = "10k" #10 kb of identifiers list per chunk

#you can specify that entries could be as big as (only used for caching files for remote extraction)
#if an entry is requested that is bigger than this specified maximum, the cache file will be
#downloaded again
MAXENTRYSIZE = 204800 #200kb

# for remote compressed flatfiles we apply a conservative (+40%) estimation
# (from reported average compress_ratio), i.e. we download more than strictly
# necessary, since the compress_ratio is an average and the particular chunk of
# the compressed file could have different compression ratio.
# If many entries fail to be extracted, considering raising the estimation
# (e.g. 5000 for +50%, 4000 for +60%..). Conversely to save bandwidth you may
# lower it (e.g. use 9000 for only +10% overdownload amount)
OVERDLFACTOR = 6000

# for remote compressed flatfiles when using cache we assume the default 10MiB span
# between index points on the gztool uncompressed stream (governed by -s option). If
# you are using a different span, modify the following value
GZCHUNKSPAN = 10485760 #10Mib

#CONSTANTS
RECOMPRESSRATIO = re.compile(r'^\tGuessed gzip.* \(([0-9.]+)%\) .*$', re.MULTILINE)
REGZINDEX = re.compile(r'^#([0-9]+): @ ([0-9]+) / ([0-9]+) .*$')
PROGNAME = "extractor.py"
VERSION = "3.3"
AUTHOR = "Giuseppe Insana"
args = None


def unzip(iterable):
    """
    reverse of zip
    """
    return zip(*iterable)


def batch(iterable, batchsize=1):
    """
    split iterable in constant-size chunks
    """
    length = len(iterable)
    for index in range(0, length, batchsize):
        yield iterable[index:min(index + batchsize, length)]


def find_chunkstart_size(startpos, entrysize):
    """
    use u2c_map to calculate whence to dl from
    and use compressratio and entrysize
    to estimate how much needed to dl
    if cache used, check if file already exists otherwise setup to dl whole chunk
    """
    gzchunk_file = None #only used if keepcache
    #print_u2c_map() #debug
    nearest_index = index_binary_search(1, args.u2c_maxindex, startpos)
    if nearest_index is None:
        eprint("    => ERROR: Could not extract correct uncompressed/compressed byte mapping")
        sys.exit(5)
    uncompressed_chunkstart = args.u2c_index[nearest_index]
    chunkstart = max(0, args.u2c_map[uncompressed_chunkstart] - 1) #1 byte before index point start
    if args.keepcache:
        gzchunk_file = args.gzchunks_fprefix + "." + str(nearest_index)
        # check if gzchunk already exists in cache
        if os.path.isfile(gzchunk_file) and entrysize <= args.maxentrysize:
            chunksize = 0 #do not download anything, use previously downloaded cached file
        else: # if it doesn't, download a whole gzchunk and store it in cache
            if entrysize > args.maxentrysize:
                if args.verbose:
                    eprint("    => NOTICE: entry size {}, consider increasing MAXENTRYSIZE".format(
                        entrysize))
                args.maxentrysize = entrysize #push up the boundary since we encountered bigger one
            chunksize = GZCHUNKSPAN * args.compress_ratio // OVERDLFACTOR + args.maxentrysize
    else:
        chunksize = (startpos - uncompressed_chunkstart) \
            * args.compress_ratio // OVERDLFACTOR + entrysize
        #NOTE: we add the whole entrysize without applying compress_ratio (avg) as we assume
        #the worst, i.e. we assume worst compression for each single entry

    #if args.verbose: #debug
    #   eprint(" |-- found nearest lower index is {}: {} ---> chunkstart {} ; chunksize {}".format(
    #       nearest_index, uncompressed_chunkstart, chunkstart, chunksize)) #debug
    return chunkstart, chunksize, gzchunk_file


def retrieve_compressed_chunk(chunkstart, chunksize, temp_dl_file, gzchunk_file):
    """
    create a temporary file downloading a chunk of a remote compressed flatfile
    optionally store in cache a whole gzchunk
    """
    #create temp_dl_file containing only chunkstart zeros (not taking space on disk):
    dd_call = ["dd", 'if=/dev/zero', "of={}".format(temp_dl_file),
               "seek={}".format(chunkstart), 'bs=1', 'count=0']
    #eprint(" |-- calling dd as '{}'".format(" ".join(dd_call)))
    #result = run(dd_call, check=True, capture_output=True) #subprocess.run #py3.7only
    result = run(dd_call, check=True, stdout=PIPE, stderr=PIPE) #subprocess.run #py3.6ok
    if result.returncode != 0:
        eprint("    => ERROR: problems with dd command used for compressed temporary file")
        eprint("    => {}".format(result.stderr.decode()))
        sys.exit(1)
    CHUNKFH = open("{}".format(temp_dl_file), 'ab') #append, binary
    #append downloaded chunk to temp_dl_file
    if args.keepcache and args.verbose:
        eprint(" |-- downloading {} bytes for gzchunk cache".format(chunksize)) #debug
    CHUNKFH.write(retrieve_from_size(args.flatfile, chunkstart, chunksize))
    CHUNKFH.close()
    if args.keepcache: #store the downloaded gzchunk in the cache
        #eprint(" |-- storing new gzchunk file in cache: '{}'".format(gzchunk_file)) #debug
        os.rename(temp_dl_file, gzchunk_file)


def retrieve_from_size(url, begin, size):
    """
    reads from remote url a specified amount of bytes
    """
    end = begin + size - 1
    headers = {'user-agent': 'ffdb/2.3.6',
               'Accept-Encoding': 'identity',
               'range': 'bytes={}-{}'.format(begin, end)}

    if url[0:3] == "ftp":
        eprint("    => ERROR: ftp scheme not supported for range retrieval")
        sys.exit(5)
    else: #http
        r = get(url, headers=headers)
    #eprint("request headers: {}".format(r.request.headers)) #debug
    #eprint("response headers: {}".format(r.headers)) #debug
    if r.status_code in (200, 206):
        return r.content
    eprint("    => ERROR: problems retrieving entry, http status code is '{}' ({})".format(
        r.status_code, r.reason))
    return None


def find_compressratio():
    """
    find how much data was compressed
    """
    gztool_call = [GZTOOL_EXE, '-l', args.gzindex]
    result = run(gztool_call, check=True, stdout=PIPE, stderr=PIPE) #subprocess.run
    if result.returncode != 0:
        eprint("    => ERROR: problem using gztool to read compressed_index {}".format(
            args.gzindex))
        eprint("    => {}".format(result.stderr.decode()))
        sys.exit(5)
    compressratio_match = RECOMPRESSRATIO.search(result.stdout.decode())
    if compressratio_match:
        reduced_ratio = compressratio_match.group(1)
        args.compress_ratio = int(10000 - float(reduced_ratio) * 100)
        if args.verbose:
            eprint(" |-- the flatfile is compressed to {}% of the original".format(
                round(100 - float(reduced_ratio), 2)))
    else:
        eprint("    => ERROR: problem using gztool to find compress_ratio")
        eprint("    => {}".format(result.stdout.decode()))
        eprint("    => WARNING: could not deduce correct compressratio from index file,")
        eprint("                for safety we'll download more than strictly necessary")
        args.compress_ratio = 10000


def print_u2c_map():
    """
    print uncompressed to compressed index mapping (debug)
    """
    eprint(" |-- map's max index is {}".format(args.u2c_maxindex))
    for index_point in args.u2c_index:
        eprint(" |-- {}: {} ---> {}".format(index_point, args.u2c_index[index_point],
                                            args.u2c_map[args.u2c_index[index_point]]))


def build_u2c_map():
    """
    to build uncompressed to compressed index mapping
    """
    u2c_index = dict()
    u2c_map = dict()
    gztool_call = [GZTOOL_EXE, '-ll', args.gzindex]
    if args.verbose:
        eprint(" |-- building gztool index: '{}'".format(" ".join(gztool_call)))
    gztool_pipe = Popen(gztool_call, stdout=PIPE,
                        bufsize=1, universal_newlines=True, stderr=DEVNULL)
    for line in gztool_pipe.stdout:
        for index_string in line.split(", "):
            gzindex_match = REGZINDEX.match(index_string)
            if gzindex_match:
                index_point, compressed_byte, uncompressed_byte = gzindex_match.groups()
                u2c_index[int(index_point)] = int(uncompressed_byte)
                u2c_map[int(uncompressed_byte)] = int(compressed_byte)

    args.u2c_maxindex = len(u2c_index)
    args.u2c_index = u2c_index
    args.u2c_map = u2c_map


def check_args():
    """
    parse arguments and check for error conditions
    """
    global args, GZTOOL_EXE
    usagetxt = """{0} -f FLATFILE -i INDEXFILE -s IDENTIFIER
                          or
       {0} -f FLATFILE -i INDEXFILE -l LISTFILE
    [-s] : one or more identifiers (space separated)
    [-l] : filename containing a list of identifiers
    see '{0} -h' for tweaks and optional modes
    \nnotes: FLATFILE can be local or remote (use URL as FLATFILE), it can contain
         encrypted or compressed entries, it can even be gzipped as a whole.
         All types will be handled automatically.
       For gzipped FLATFILE, the gztool exe and a GZINDEX need to be available
         see {0} -h for tweaks and optional modes\nexamples:
    {0} -f entries.dat -i entries.pos -s Q9GJU7 Q8HYU5 >twoentries.dat
      (extract the entries identified by Q9GJU7 and Q8HYU5)
    {0} -f entries.dat -i entries.pos -d -s 9606 >homo.dat
      (extract all entries corresponding to identifier 9606)

    {0} -f entries.dat -i entries.pos -l chosen.ids >chosen.dat
    {0} -f entries.dat.gz -i entries.pos -l chosen.ids -o chosen.dat

    {0} -f http://hostname/db -i db.pos -s duffyduck
    {0} -f http://hostname/db.gz -i db.pos -I db.gzi -s duffyduck
      (from a remote db extract entry 'duffyduck': -I needed if gzipped)
    """.format(PROGNAME)
    parser = argparse.ArgumentParser(description='Use a positional index to retrieve \
                                     entries from a flatfile', usage=usagetxt)
    parser.add_argument('-f', '--file', dest='flatfile',
                        help="filename of flatfile to be processed; can be an URL",
                        required=True, type=str)
    parser.add_argument('-i', '--index', dest='index_filename',
                        help="filename of index file with entry identifiers",
                        required=True, type=str)
    parser.add_argument('-s', '--single', dest='identifiers',
                        help="identifier(s) for the desired entry to be extracted",
                        required=False, type=str, nargs='+')
    parser.add_argument('-t', '--threads', dest='threads',
                        help="use specified number of multiple threads for parallel retrieval. \
                        See also -b option.", required=False, type=int)
    parser.add_argument('-l', '--list', dest='list_filename',
                        help="a file containing a list of identifiers for entries to retrieve",
                        required=False, type=str)
    parser.add_argument('-o', '--outfile', dest='output_filename',
                        help="optionally write output to file rather than to stdout",
                        required=False, type=str)
    parser.add_argument('-m', '--mergedretrieval', dest='merged', action='store_true',
                        help="merge and retrieve together adjacent entries (requires more memory \
                        and processing but could prove much faster for remote extraction)",
                        required=False)
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help="verbose operation", required=False)
    parser.add_argument('-d', '--duplicates', dest='duplicates', action='store_true',
                        help="specify INDEX_FILE could contain duplicate identifiers and request \
                        extraction of all of them (default is to extract a single entry)",
                        required=False)
    parser.add_argument('-z', '--zfound', dest='zfound', action='store_true',
                        help="specify INDEX_FILE contains duplicate identifiers and request \
                        extraction of last entry appearing in the flatfile (default is the first)",
                        required=False)
    parser.add_argument('-x', '--xsanity', dest='xsanity', action='store_true',
                        help="check checksums (if provided in index) to confirm sanity of \
                        the extracted data (slower extraction)", required=False)
    parser.add_argument('-p', '--passphrase', dest='passphrase',
                        help="specify passphrase to decrypt entries from encrypted flatfile; \
                        not recommended on multiuser systems due to security concerns. By default \
                        the passphrase will be requested interactively", required=False, type=str)
    parser.add_argument('-b', '--blocksize', dest='list_blocksize',
                        help="define blocksize (size of LIST_FILENAME chunks) for parallel \
                        extraction. This is fast and has lowest memory requirements, but could \
                        be less efficient if lots of entries are mapped to same identifier when \
                        --duplicates is used. If unspecified, it will be adjusted automatically \
                        to number of threads. Use -b 0 to disable block extraction",
                        required=False, type=siprefix2num)
    parser.add_argument('-c', '--compressed', dest='compressed', action='store_true',
                        help="specify flatfile is gzipped; a .gzi GZINDEX file is required",
                        required=False)
    parser.add_argument('-g', '--gzindex', dest='gzindex',
                        help="filename of the compressed index .gzi",
                        required=False, type=str)
    parser.add_argument('-r', '--remote', dest='remote', action='store_true',
                        help="specify flatfile is remote and treat FLATFILE as an URL",
                        required=False)
    parser.add_argument('-k', '--keepcache', dest='keepcache', action='store_true',
                        help="keep local cache to consume less bandwidth; only applicable to \
                        remote gzipped flatfiles", required=False)
    args = parser.parse_args()

    randnum = str(randint(1000, 9999))

    args.progressbar = False
    if args.verbose:
        eprint(" .-- {} v{} -- by {} --.".format(PROGNAME, VERSION, AUTHOR))
        args.progressbar = True

    if args.zfound and args.duplicates:
        eprint("    => ERROR: No sense specifying both --zfound and --duplicates at the same time")

    if args.verbose:
        if args.zfound:
            eprint(" |-- in case of duplicate ids, the last entry in ff will be extracted")
        elif args.duplicates:
            eprint(" |-- in case of duplicate ids, all entries in ff will be extracted")
            if args.list_blocksize is not None:
                eprint("    => NOTICE: -b not recommended if many entries with same identifier requested")

    if args.flatfile[-3:] == ".gz" and not args.compressed:
        eprint("    => NOTICE: -f argument has extension .gz: assuming flatfile is compressed")
        args.compressed = True

    if args.flatfile.find("://") != -1 and not args.remote:
        eprint("    => NOTICE: -f argument appears to be an URL: assuming flatfile is remote")
        args.remote = True

    if not args.remote:
        args.flatfilesize = os.path.getsize(args.flatfile) #find and store local flatfile filesize

    if args.identifiers is None and args.list_filename is None:
        eprint("    => ERROR: At least one of --list or --single has to be specified!")
        sys.exit(22)

    if args.identifiers is not None and args.list_filename is not None:
        eprint("    => ERROR: No sense specifying both --list and --single at the same time")
        sys.exit(22)

    if args.list_blocksize is not None:
        if args.threads is None:
            eprint("    => ERROR: specifying blocksize makes sense only for -t execution")
            sys.exit(22)
        if args.identifiers is not None:
            eprint("    => ERROR: No sense specifying --blockextraction with --single identifiers")
            sys.exit(22)
        #if set to 0 it means blockextraction is disabled

    if args.merged:
        if args.list_blocksize is not None:
            eprint("    => NOTICE: --blocksize cannot be used with --mergedretrieval, ignoring -b")
        args.list_blocksize = 0 #disable

    if args.threads is not None: #multithread
        if args.threads < 2:
            eprint("    => ERROR: No sense specifying a number of threads lower than 2!")
            sys.exit(22)
        args.mt_subfiles_prefix = TEMPDIR + "/tmpEXTRACTmts" + randnum
        if args.list_filename:
            if args.list_blocksize is None: #if not specified, we compute it
                args.list_blocksize = max(calculate_blocksize(args.list_filename, args.threads),
                                          siprefix2num(MINBLOCKSIZE))
        else: #single identifiers
            args.list_blocksize = 0

        if args.list_blocksize != 0: #if not disabled or incompatible
            #we'll use blockextraction without prior collection
            args.list_filesize, args.chunks_count = calculate_chunknum(
                args.list_filename, args.list_blocksize)
            if args.verbose:
                eprint(" |-- parallel work on {} chunks of maximum {} bytes (-b to change)".format(
                    args.chunks_count, args.list_blocksize))
        else: #either mergedretrieval or parallel without blocksize (we'll first collect indexes)
            args.chunks_count = args.threads
        if args.verbose:
            eprint(" |-- using maximum {} parallel threads (-t); your OS reports {} cpus.".format(
                args.threads, os.cpu_count()))
    else: # if unspecified, set args.threads to 1
        args.threads = 1
        args.chunks_count = 1

    if args.verbose and args.merged:
        eprint(" |-- using merged retrieval for adjacent entries")

    args.decrypt = False
    args.deflated = False

    #gather information from first line of index
    with open(args.index_filename, 'r', 1) as indexfh:
        index_type, args.cipher_name, args.keysize, has_checksum = check_index(
            indexfh.readline()) #pass first line of index as argument
    if args.xsanity and not has_checksum:
        eprint("    => ERROR: sanity check requested (-x) but index does not contain checksums!")
        sys.exit(1)
    if args.xsanity and args.verbose:
        eprint(" |-- entry checksums will be verified")
    if index_type in (".", "+"):
        args.decrypt = True
        if args.verbose:
            eprint(" |-- index made for encrypted entries")
    if index_type in (":", "+"):
        args.deflated = True
        if args.verbose:
            eprint(" |-- index made for compressed entries")

    if args.decrypt:
        if args.passphrase is None:
            eprint(" |-- entries are encrypted, please provide a passphrase:")
            args.passphrase = getpass.getpass(prompt=" |>> Passphrase: ")
        my_cipher_name, args.key = derive_key(args.passphrase, args.keysize)
        if my_cipher_name != args.cipher_name:
            eprint("    => ERROR: problems with cipher name (this should not have happened!)")
            sys.exit(14)
        if args.verbose:
            eprint(" |-- encrypted flatfile (with cipher {}) will be processed".format(
                args.cipher_name))
            #eprint(" |-- the passphrase is: {}".format(args.passphrase)) #DEBUG!
            #eprint(" |-- the decryption key is: {}".format(args.key)) #DEBUG!
    else: #not encrypted
        if args.passphrase is not None:
            eprint("    => NOTICE: ignoring specified passphrase:")
            eprint("       index was not made for encrypted flatfile!")

    if args.compressed:
        if args.decrypt:
            eprint("    => ERROR: compressed flatfile and encrypted flatfile")
            eprint("       are mutually exclusive.")
            eprint("       If flatfile was compressed during encryption")
            eprint("       this will be handled automatically")
            sys.exit(22)
        if GZTOOL_EXE is None:
            result = run(["which", "gztool"], check=True, stdout=PIPE, stderr=PIPE) #subprocess.run
            if result.returncode != 0:
                eprint("    => ERROR: gztool command not found in your path.")
                eprint("Extraction of entries from compressed file depends on it.")
                eprint("Please install gztool to continue: https://github.com/circulosmeos/gztool")
                eprint("    => {}".format(result.stderr.decode()))
                sys.exit(2)
            GZTOOL_EXE = result.stdout.decode().rstrip()
        if args.gzindex is None:
            if args.remote:
                eprint("    => ERROR: a local compressed index needs to be specified (-g)!")
                sys.exit(22)
            else:
                args.gzindex = args.flatfile + "i" #default when not specified
                if args.verbose:
                    eprint("    => NOTICE: assuming COMPRESSEDINDEX is '{}'.".format(args.gzindex))
                    eprint("       Use -I option to override")
            if args.verbose:
                eprint(" |-- building hash of compressed index points..")
        if args.remote:
            args.dlfiles_prefix = TEMPDIR + "/tmpEXTRACTdl" + randnum

    if args.keepcache:
        if not args.remote or not args.compressed:
            eprint("    => ERROR: --keepcache only applicable for remote compressed flatfiles")
            sys.exit(22)
        args.maxentrysize = MAXENTRYSIZE #boundary can be increased if a bigger entry is requested
        args.remote_name = args.flatfile.split("/")[-1]
        args.gzcache_dir = TEMPDIR + "/tmpEXTRACTcache/" + args.remote_name + "/"
        args.gzchunks_fprefix = args.gzcache_dir + "GZ"
        if args.verbose:
            eprint(" |-- using cache of compressed chunks at '{}'".format(args.gzcache_dir))
            eprint(" |-- (you can use environment variable TMPDIR to change this location)")


def check_files():
    """
    do some checks on availability of resources
    """
    if args.remote:
        try:
            retrieve_from_size(args.flatfile, 0, 16)
        except FileNotFoundError:
            eprint("    => ERROR: Cannot access remote url '{}'".format(args.flatfile))
            sys.exit(2)
    else:
        check_iofiles([args.flatfile], [])
    if args.compressed:
        check_iofiles([args.gzindex], [])
        find_compressratio() #find how much data was compressed
        build_u2c_map() #to build uncompressed to compressed index mapping
    if args.list_filename is not None:
        check_iofiles([args.list_filename], [])
    if args.output_filename is not None:
        check_iofiles([], [args.output_filename])

    if args.threads > 1: #multithread
        args.chunk_tempfh = list()
        args.chunk_tempfiles = list()
        for chunknum in range(args.chunks_count):
            chunknumstr = str(chunknum).zfill(len(str(args.chunks_count))) #e.g. 00, 01..
            chunk_tempfile = args.mt_subfiles_prefix + "." + chunknumstr
            try:
                myoutputfh = open(chunk_tempfile, 'wb')
            except PermissionError:
                eprint("    => ERROR: Cannot open temporary file '{}' for writing".format(
                    chunk_tempfile))
                sys.exit(1)
            args.chunk_tempfh.append(myoutputfh) #store temp filehandles
            args.chunk_tempfiles.append(chunk_tempfile) #store temp filenames

    if args.remote and args.compressed:
        args.chunk_dl_tempfiles = list()
        for chunknum in range(args.chunks_count): #if singlethread, only one
            chunknumstr = str(chunknum).zfill(len(str(args.chunks_count))) #e.g. 00, 01..
            chunk_tempfile = args.dlfiles_prefix + "." + chunknumstr
            try:
                myoutputfh = open(chunk_tempfile, 'wb')
            except PermissionError:
                eprint("    => ERROR: Cannot open temporary file '{}' for writing".format(
                    chunk_tempfile))
                sys.exit(1)
            args.chunk_dl_tempfiles.append(chunk_tempfile)
        if args.keepcache:
            if not os.path.isdir(args.gzcache_dir): #create dir if not already present
                os.makedirs(args.gzcache_dir, mode=0o700, exist_ok=False)
            chunk_tempfile = args.gzchunks_fprefix + ".0"
            try:
                myoutputfh = open(chunk_tempfile, 'wb')
            except PermissionError:
                eprint("    => ERROR: Cannot open temporary cache file '{}' for writing".format(
                    chunk_tempfile))
                sys.exit(1)


def print_entry(entry, identifier, fh, chunknum, merged=None):
    """
    apply eventual post-processing and then print entry to filehandle
    """
    if merged is not None:
        #eprint("got merged: {} of size {}".format(merged, len(entry['content']))) #debug
        for collated in merged:
            subentry = dict()
            #unpack collated information
            if args.xsanity:
                if args.decrypt:
                    position, entry_length, identifier, subentry['iv'], \
                        subentry['checksum'] = collated
                else:
                    position, entry_length, identifier, subentry['checksum'] = collated
            elif args.decrypt:
                position, entry_length, identifier, subentry['iv'] = collated
                #eprint("got subentry {}: {}-{} {}".format(identifier,
                #                                         position, entry_length,
                #                                         subentry['iv'])) #debug
            else:
                position, entry_length, identifier = collated
                #eprint("got subentry {}: {}-{}".format(identifier, position, entry_length)) #debug
            subentry['content'] = entry['content'][position:position + entry_length]
            #now call again print_entry for each entry making up the merged entry
            print_entry(subentry, identifier, fh, chunknum)
    else:
        entryenc = None
        #eprint("printing entry for {} of size {}".format(identifier, len(entry['content']))) #debug
        if args.decrypt:
            iv = b''.fromhex(entry['iv']) #convert from hex to bytes
            cipher = init_cipher(args.key, iv)
            if args.deflated:
                try:
                    entryenc = inflate(cipher.decrypt(entry['content']))
                except IOError:
                    entryenc = None
                    corrupted_count[chunknum] += 1
                    if args.verbose:
                        eprint("    => WARNING: could not decrypt entry for identifier {}".format(
                            identifier))
            else:
                try:
                    entryenc = cipher.decrypt(entry['content'])
                except IOError:
                    entryenc = None
                    corrupted_count[chunknum] += 1
                    if args.verbose:
                        eprint("    => WARNING: could not decrypt or inflate entry")
                        eprint("       for identifier {}".format(identifier))
        elif args.deflated:
            try:
                entryenc = inflate(entry['content'])
            except IOError:
                entryenc = None
                corrupted_count[chunknum] += 1
                if args.verbose:
                    eprint("    => WARNING: could not inflate entry")
                    eprint("       for identifier {}".format(identifier))
        else:
            entryenc = entry['content']

        if entryenc:
            if args.xsanity:
                entry_checksum = int_to_b64(zlib.crc32(entryenc))
                if entry_checksum != entry['checksum']:
                    if args.verbose:
                        eprint("    => WARNING: Skipping entry due to failed checksum sanity check")
                        eprint("                for identifier {}".format(identifier))
                    corrupted_count[chunknum] += 1
                    return
            extracted_count[chunknum] += 1
            if fh.mode == 'w':
                fh.write(entryenc.decode('UTF-8'))
            else:
                fh.write(entryenc)
            fh.flush()
        else:
            if args.verbose:
                eprint("    => WARNING: skipping empty entry for '{}'".format(identifier))
    return


def index_binary_search(min_index, max_index, search_value):
    """
    perform a binary search inside the gzindex of compressed/uncompressed points
    """
    u2c_index = args.u2c_index
    found_index = None
    #eprint("searching for {}".format(search_value)) #debug
    while min_index <= max_index:
        mid_index = (min_index + max_index + 1) // 2
        #eprint("      min: {} max: {} found: {} mid: {}".format(
        #  min_index, max_index, found_index, mid_index))
        if u2c_index[mid_index] <= search_value:
            min_index = mid_index + 1
            found_index = mid_index
        else:
            max_index = mid_index - 1
    return found_index


def fetch_compressed_entry(flatfile, position, entry_length):
    """
    use gztool to extract a compressed_entry
    from given position of specified flatfile
    """
    gztool_call = [GZTOOL_EXE, '-W', '-b', str(position), '-I', args.gzindex, flatfile]
    #eprint(" |-- calling gztool as '{}'".format(" ".join(gztool_call))) #debug

    with Popen(gztool_call, stdout=PIPE, stderr=DEVNULL) as gztool_pipe:
        return gztool_pipe.stdout.read(entry_length)


def collect_index(indexfh, identifier):
    """
    find and decode position of entries corresponding to specified identifier from index
    return zipped arrays of decoded positions, sizes and optionally iv and checksums
    """
    if args.duplicates:
        if args.xsanity:
            positions, checksums = get_positions_checksums(indexfh, identifier)
        else:
            positions = get_positions(indexfh, identifier)
    elif args.zfound:
        if args.xsanity:
            position, checksum = get_position_checksum_last(indexfh, identifier)
            positions = [position]
            checksums = [checksum]
        else:
            positions = [get_position_last(indexfh, identifier)]
    else:
        if args.xsanity:
            position, checksum = get_position_checksum_first(indexfh, identifier)
            positions = [position]
            checksums = [checksum]
        else:
            positions = [get_position_first(indexfh, identifier)]

    if not positions or positions == [None]: #empty array
        if args.verbose:
            eprint("    => WARNING: '{}' not found in index; skipping".format(identifier))
        return []

    found_positions = list()
    found_lengths = list()
    if args.decrypt:
        found_iv = list() #to store iv

    found_count[0] += len(positions)

    identifiers = [identifier] * len(positions) #to identify entries failing to extract

    for position in positions:
        iv = ''
        checksum = ''
        #1) extract position and entry size
        if args.decrypt: #with iv
            posmatch = REESIV.match(position)
            position, entry_length, iv = posmatch.groups()
            found_iv.append(iv)
        else: #no iv
            posmatch = REES.match(position)
            position, entry_length = posmatch.groups()

        found_positions.append(b64_to_int(position)) #integer position
        found_lengths.append(b64_to_int(entry_length)) #integer entry size

    if args.xsanity:
        if args.decrypt:
            return zip(found_positions, found_lengths, identifiers, found_iv, checksums)
        else:
            return zip(found_positions, found_lengths, identifiers, checksums)
    elif args.decrypt:
        return zip(found_positions, found_lengths, identifiers, found_iv)
    else:
        return zip(found_positions, found_lengths, identifiers)


def merge_adjacent(found_indexes):
    """
    find if adjacent entries requested and group them together before extraction
    strategy:
    check whether next entry follows previous.
     if yes: add to a merged entry with same position, sum of entry_sizes,
     collation of remaining data and keeping original entries' size and pos for
     later unpacking by print_entry
     otherwise: keep as normal (isolated) entry
    """
    merged_indexes = list()
    prev_position = prev_length = 0.1 #bogus for first comparison
    prev_rest = []
    merged_length = 0
    merged_position = None
    merged_collated = list()
    for position, entry_length, *rest in found_indexes:
        if prev_position + prev_length == position:
            #eprint('{}-{} is adjacent to {}-{}, MERGING'.format(position, entry_length,
            #                                                   prev_position, prev_length))
            merged_length += prev_length
            if merged_position is None:
                merged_position = prev_position
            merged_collated.append((prev_position - merged_position, prev_length, *prev_rest))
        else:
            #eprint('{}-{} NOT adjacent to {}-{}'.format(position, entry_length,
            #                                           prev_position, prev_length))
            if merged_length:
                merged_length += prev_length
                merged_collated.append((prev_position - merged_position, prev_length, *prev_rest))
                merged_indexes.append((merged_position, merged_length, merged_collated))
                merged_length = 0 #reset
                merged_position = None #reset
                merged_collated = []
            else:
                merged_indexes.append((prev_position, prev_length, *prev_rest))
        prev_position = position
        prev_length = entry_length
        prev_rest = rest
    if merged_length: # if we arrive to the end with an opened merged entry
        #eprint("Final append of merged entry for {}-{}".format(position, entry_length)) #debug
        merged_length += entry_length
        merged_collated.append((position - merged_position, entry_length, *prev_rest))
        merged_indexes.append((merged_position, merged_length, merged_collated))
    else: # last isolated entry
        #eprint("Final append of isolated entry for {}-{}".format(position, entry_length)) #debug
        merged_indexes.append((position, entry_length, *prev_rest))

    merged_indexes.pop(0) #remove bogus first element

    return merged_indexes


def extract_entry(flatfilefh, indexfh, identifier, chunknum):
    """
    find position of entry corresponding to specified identifier from index
    calls retrieve_entry (which calls print_entry)
    """
    if args.duplicates:
        if args.xsanity:
            positions, checksums = get_positions_checksums(indexfh, identifier)
        else:
            positions = get_positions(indexfh, identifier)
    elif args.zfound:
        if args.xsanity:
            position, checksum = get_position_checksum_last(indexfh, identifier)
            positions = [position]
            checksums = [checksum]
        else:
            positions = [get_position_last(indexfh, identifier)]
    else:
        if args.xsanity:
            position, checksum = get_position_checksum_first(indexfh, identifier)
            positions = [position]
            checksums = [checksum]
        else:
            positions = [get_position_first(indexfh, identifier)]

    if not positions or positions == [None]: #empty array
        if args.verbose:
            eprint("    => WARNING: '{}' not found in index; skipping".format(identifier))
        return

    found_count[chunknum] += len(positions)
    iv = None
    checksum = None
    for position in positions:
        #1) extract position and entry size
        if args.decrypt: #with iv
            posmatch = REESIV.match(position)
            position, entry_length, iv = posmatch.groups()
        else: #no iv
            posmatch = REES.match(position)
            position, entry_length = posmatch.groups()
        if args.xsanity:
            checksum = checksums.pop(0)

        #2) decode position
        position = b64_to_int(position)
        entry_length = b64_to_int(entry_length)

        #if args.verbose:
        #  eprint(" |-- #{} extracting '{}' from position '{}' length '{}'".format(
        #      chunknum, identifier, position, entry_length)) #debug

        #3) retrieve entry
        retrieve_entry(flatfilefh, identifier, position, entry_length, chunknum, iv=iv,
                       checksum=checksum)


def retrieve_entry(flatfilefh, identifier, position, entry_length, chunknum, iv=None, checksum=None,
                   merged=None):
    """
    Note: for remote flatfiles, flatfilefh is ignored, using instead args.flatfile (the url)
    retrieve the entry from the flatfile
    calls print_entry
    """
    entry = dict()
    entry['iv'] = iv #if given
    entry['checksum'] = checksum #if given
    #eprint("retrieving {}: {}-{}".format(identifier, position, entry_length)) #debug
    if args.compressed: #gzipped flatfile
        if args.remote:
            chunkstart, chunksize, gzchunk_file = find_chunkstart_size(position,
                                                                       entrysize=entry_length)
            if chunksize != 0: #cache file already exists, no need to retrieve
                retrieve_compressed_chunk(chunkstart, chunksize,
                                          args.chunk_dl_tempfiles[chunknum], gzchunk_file)
            #else: #debug
            #   if args.verbose:
            #       eprint(" |-- using cached gzchunk file: '{}'".format(gzchunk_file)) #debug
            if args.keepcache:

                entry['content'] = fetch_compressed_entry(gzchunk_file, position, entry_length)
            else:
                entry['content'] = fetch_compressed_entry(args.chunk_dl_tempfiles[chunknum],
                                                          position, entry_length)
        else:
            entry['content'] = fetch_compressed_entry(args.flatfile,
                                                      position, entry_length)
    else: #not gzipped
        if args.remote:
            entry['content'] = retrieve_from_size(args.flatfile, #url
                                                  position,
                                                  entry_length)
        else:
            entry['content'] = read_from_size(flatfilefh, position, entry_length)
    if args.threads == 1: #final printout
        print_entry(entry, identifier, args.outputfh, chunknum, merged=merged)
    else: #print to separate temp files
        print_entry(entry, identifier, args.chunk_tempfh[chunknum], chunknum, merged=merged)


def collect_extract_entries():
    """
    identify entries to extract and collect them, optionally merge together, then extract them
    either single or multithreaded
    """
    global requested_count
    indexfh = open(args.index_filename, 'r', 1)

    if args.merged:
        found_indexes = SortedList() #sorted by integer position
        if args.list_filename:
            with open(args.list_filename) as listfh:
                for line in listfh:
                    found_indexes.update(collect_index(indexfh, line.rstrip()))
                    requested_count[0] += 1
        else:
            for identifier in args.identifiers:
                found_indexes.update(collect_index(indexfh, identifier))
                requested_count[0] += 1
    else:
        found_indexes = list()
        if args.list_filename:
            with open(args.list_filename) as listfh:
                for line in listfh:
                    found_indexes.extend(collect_index(indexfh, line.rstrip()))
                    requested_count[0] += 1
        else:
            for identifier in args.identifiers:
                found_indexes.extend(collect_index(indexfh, identifier))
                requested_count[0] += 1

    indexfh.close()

    #eprint(" found_indexes: {}".format(list(found_indexes))) #debug

    #merge adjacent entries together for faster extraction
    found_length = len(found_indexes)
    if args.merged:
        indexes = merge_adjacent(found_indexes)
        del found_indexes
    else:
        indexes = found_indexes

    #eprint(" indexes: {}".format(indexes)) #debug

    #use positional information from possibly merged indexes to retrieve and print entries
    if indexes: #if not empty
        if args.verbose:
            if args.merged:
                merged_length = len(indexes)
                if merged_length < found_length:
                    eprint(" |-- {} retrievals for {} entries ({} merged)".format(
                        merged_length, found_length, found_length - merged_length))
        if args.threads > 1: #multithread for merged strategy: submit indexes to threads in batches
            indexes_length = len(indexes)
            batchsize = ceil(indexes_length / args.threads)
            batchesnum = ceil(indexes_length / batchsize)

            if args.verbose:
                if args.merged:
                    eprint(" |-- parallel work on {} batches of maximum {} retrievals".format(
                        batchesnum, batchsize))
                else:
                    eprint(" |-- parallel work on {} batches of maximum {} entries".format(
                        batchesnum, batchsize))
            #eprint("batched indexes: {}".format(list(batch(indexes, batchsize)))) #debug

            #init threads
            pool = Pool(args.threads, initializer=init_thread_collectmode, initargs=(
                extracted_count, corrupted_count))
            if args.progressbar:
                _ = list(tqdm(pool.imap(retrieve_entries, batch(indexes, batchsize)),
                              total=batchesnum, ascii=PROGRESSBARCHARS))
            else:
                pool.imap(retrieve_entries, batch(indexes, batchsize))
            pool.close() #no more work to submit
            pool.join() #wait workers to finish
        else:
            retrieve_entries(indexes)


def retrieve_entries(indexes):
    """
    use the (now merged where possible) indexes to retrieve entries from flatfile
    """
    if args.remote:
        flatfilefh = None
    else:
        flatfilefh = open(args.flatfile, 'rb')

    if current_process().name == 'MainProcess':
        chunknum = 0
    else: #for debug of multithreading
        chunknum = int(format(current_process().name.split('-')[1]))-1 #-1 to start with 0

    #eprint(" [{}] processing: {}".format(chunknum, indexes)) #debug
    #found_positions, found_lengths, found_identifiers = unzip(indexes) #debug
    #eprint(" [{}] found_positions: {}".format(chunknum, found_positions)) #debug
    #eprint(" [{}] found_lengths: {}".format(chunknum, found_lengths)) #debug
    for position, entry_length, *rest in indexes:
        if isinstance(rest[0], list):
            #eprint("{}-{} is merged entry".format(position, entry_length)) #debug
            retrieve_entry(flatfilefh, 'merged', position, entry_length, chunknum, merged=rest[0])
        else:
            #eprint("{}-{} is normal entry: {}".format(position, entry_length, rest[0])) #debug
            if args.xsanity:
                if args.decrypt:
                    retrieve_entry(flatfilefh, rest[0], position, entry_length, chunknum,
                                   iv=rest[1], checksum=rest[2])
                else:
                    retrieve_entry(flatfilefh, rest[0], position, entry_length, chunknum,
                                   checksum=rest[1])
            elif args.decrypt:
                retrieve_entry(flatfilefh, rest[0], position, entry_length, chunknum, iv=rest[1])
            else:
                retrieve_entry(flatfilefh, rest[0], position, entry_length, chunknum)

    if flatfilefh is not None:
        flatfilefh.close()


def extract_entries(chunknum):
    """
    identify entries to extract and extract them
    """
    global requested_count
    indexfh = open(args.index_filename, 'r', 1)
    if args.remote:
        flatfilefh = None
    else:
        flatfilefh = open(args.flatfile, 'rb')

    if args.list_filename:
        with open(args.list_filename) as listfh:
            if args.threads == 1: #single thread
                for line in listfh:
                    extract_entry(flatfilefh, indexfh, line.rstrip(), chunknum)
                    requested_count[chunknum] += 1
            else: #multithread
                for line in chunk_of_lines(listfh, args.list_filesize,
                                           args.list_blocksize, chunknum):
                    extract_entry(flatfilefh, indexfh, line.rstrip(), chunknum)
                    requested_count[chunknum] += 1
    else:
        for identifier in args.identifiers:
            extract_entry(flatfilefh, indexfh, identifier, chunknum)
            requested_count[chunknum] += 1

    indexfh.close()
    if flatfilefh is not None:
        flatfilefh.close()


def print_stats(start_time):
    """
    if verbose print some final statistics on entries extracted (or failed to extraxct)
    """
    found_sum = 0
    requested_sum = 0
    extracted_sum = 0
    corrupted_sum = 0
    #eprint("requested: {} found: {} extracted: {} corrupted: {}".format(list(requested_count),
    #                                                                   list(found_count), #debug
    #                                                                   list(extracted_count),
    #                                                                   list(corrupted_count)))
    for chunknum in range(args.chunks_count):
        requested_sum += requested_count[chunknum]
        found_sum += found_count[chunknum]
        extracted_sum += extracted_count[chunknum]
        corrupted_sum += corrupted_count[chunknum]

    if found_sum == 1:
        eprint(" |-- Found 1 entry in the index.")
    elif found_sum > 0:
        eprint(" |-- Found {} entries in the index.".format(
            found_sum))
    if found_sum < requested_sum:
        if found_sum == 0:
            eprint("    => WARNING: NONE of the {} requested identifiers found in index!".format(
                requested_sum))
        else:
            eprint("    => WARNING: only {} of the {} requested identifiers found in index.".format(
                found_sum, requested_sum))
    if args.xsanity:
        if corrupted_sum > 0:
            eprint("    => WARNING: {} malformed entries were skipped.".format(corrupted_sum))
    if extracted_sum < found_sum:
        if extracted_sum == 0:
            eprint("    => WARNING: NONE of the {} found identifiers extracted".format(found_sum))
            eprint("       from the flatfile.")
        else:
            eprint("    => WARNING: only {} of the {} found identifiers extracted".format(
                extracted_sum, found_sum))
            eprint("       from the flatfile.")
    eprint(" '-- Elapsed: {}, {} entries/sec --'".format(*elapsed_time(start_time, extracted_sum)))


def init_thread_blockmode(r, f, x, c):
    """
    to initialise multithreaded worker
    """
    global requested_count, found_count, extracted_count, corrupted_count, req_positions
    requested_count = r
    found_count = f
    extracted_count = x
    corrupted_count = c


def init_thread_collectmode(x, c):
    """
    to initialise multithreaded worker after indexes already collected
    for merged_adjacent strategy
    """
    global extracted_count, corrupted_count, req_positions
    extracted_count = x
    corrupted_count = c


if __name__ == '__main__':
    check_args()
    check_files()

    requested_count = Array('i', args.chunks_count)
    extracted_count = Array('i', args.chunks_count)
    corrupted_count = Array('i', args.chunks_count)
    found_count = Array('i', args.chunks_count)

    if args.output_filename is None:
        args.outputfh = sys.stdout
    else:
        args.outputfh = open(args.output_filename, 'wb')
    start_secs = time.time()

    if args.threads > 1: #multithread
        if args.list_blocksize == 0: #prior collection of indexes then extraction
            collect_extract_entries() #will start multithread after collection
        else:
            #submit chunks to threads
            #init threads
            pool = Pool(args.threads, initializer=init_thread_blockmode, initargs=(
                requested_count, found_count, extracted_count, corrupted_count))
            if args.progressbar:
                _ = list(tqdm(pool.imap(extract_entries, range(args.chunks_count)),
                              total=args.chunks_count, ascii=PROGRESSBARCHARS))
            else:
                pool.imap(extract_entries, range(args.chunks_count))
            pool.close() #no more work to submit
            pool.join() #wait workers to finish

        #final union of results:
        close_subfiles(args.chunk_tempfh)
        if args.output_filename is None: #then print to stdout
            print_subfiles(args.chunk_tempfiles)
        else:
            print_subfiles(args.chunk_tempfiles, args.outputfh)

        delete_files(args.chunk_tempfiles) #cleanup
    else: #singlethread
        if args.merged:
            collect_extract_entries()
        else:
            extract_entries(0)
    if args.output_filename is not None:
        args.outputfh.close()
    if args.remote and args.compressed:
        delete_files(args.chunk_dl_tempfiles) #cleanup of temp files
    print_stats(start_secs)



#!/usr/bin/env python3
"""
remover
objective: generalized entry remover from local flatfile, it deletes entries and reindexes

by Giuseppe Insana, 2020
"""

# pylint: disable=C0103,R0912,R0915,W0603

#IMPORTS:
import os
import sys
import time
import argparse
from random import randint
from multiprocessing import Pool, Array
from sortedcontainers import SortedList
from tqdm import tqdm #progress bar
from ffdb import eprint, check_iofiles, b64_to_int, int_to_b64, \
    print_subfiles, elapsed_time, TEMPDIR, siprefix2num, calculate_chunknum, \
    calculate_blocksize, check_index, FIELDSEPARATOR, REESIVN, REES, BUFFERSIZE, \
    get_position_first, get_position_last, get_positions, delete_files, split_file, \
    PROGRESSBARCHARS

#CUSTOMIZATIONS:
#you can modify the minimum default size of block (chunk) of index identifiers
#to work on for parallel execution
MINBLOCKSIZE = "40k" #40 kb of index identifiers per chunk

#CONSTANTS
PROGNAME = "remover.py"
VERSION = "1.2"
AUTHOR = "Giuseppe Insana"
args = None


def check_args():
    """
    parse arguments and check for error conditions
    """
    global args
    usagetxt = """{0} -f FLATFILE -i INDEXFILE -l LISTFILE [-o OUTPATH]
    [-f] : flatfile from which the entries should be removed
    [-i] : index of FLATFILE
    [-l] : file with list of identifers for the entries that should be removed
    see {0} -h for tweaks and optional modes
    \nexamples:
    {0} -f entries.dat -i entries.pos -l removeme.list
      (will create entries.dat.new and entries.pos.new)
    {0} -f entries.dat -i entries.pos -l removeme.list -o cleaned
      (will create cleaned/entries.dat.new and cleaned/entries.pos.new)
    """.format(PROGNAME)
    parser = argparse.ArgumentParser(description='Use a positional index to delete \
                                     entries from a flatfile', usage=usagetxt)
    parser.add_argument('-f', '--file', dest='flatfile',
                        help="filename of flatfile from which entries should be deleted",
                        required=True, type=str)
    parser.add_argument('-i', '--index', dest='index_filename',
                        help="filename of index file containing entry identifiers",
                        required=True, type=str)
    parser.add_argument('-l', '--list', dest='list_filename',
                        help="a file containing a list of identifiers corresponding to entries \
                        to delete", required=True, type=str)
    parser.add_argument('-o', '--outpath', dest='outpath',
                        help="write new files to specified path rather than creating \
                        new files in the same location as the original ones",
                        required=False, type=str)
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help="verbose operation", required=False)
    parser.add_argument('-d', '--duplicates', dest='duplicates', action='store_true',
                        help="specify INDEX_FILE could contain duplicate identifiers and request \
                        deletion of all of them (default is to delete a single entry)",
                        required=False)
    parser.add_argument('-z', '--zfound', dest='zfound', action='store_true',
                        help="specify INDEX_FILE contains duplicate identifiers and request \
                        deletion of last entry appearing in the flatfile (default is the first)",
                        required=False)
    parser.add_argument('-t', '--threads', dest='threads',
                        help="use specified number of multiple threads for parallel reindexing",
                        required=False, type=int)
    parser.add_argument('-b', '--blocksize', dest='index_blocksize',
                        help="redefine blocksize used for parallel execution. By default \
                        it will be adjusted automatically to the number of threads",
                        required=False, type=siprefix2num)
    args = parser.parse_args()

    randnum = str(randint(1000, 9999))

    args.progressbar = False
    if args.verbose:
        eprint(" .-- {} v{} -- by {} --.".format(PROGNAME, VERSION, AUTHOR))
        args.progressbar = True

    if args.zfound and args.duplicates:
        eprint("    => ERROR: No sense specifying both --zfound and --duplicates at the same time")

    if args.verbose:
        if args.zfound:
            eprint(" |-- [-z] option selected:")
            eprint(" |   if duplicates in index, the entry appearing last in ff will be deleted")
        elif args.duplicates:
            eprint(" |-- [-d] option selected:")
            eprint(" |-- if duplicates in index, all corresponding entries in ff will be deleted")
        else:
            eprint(" |-- if duplicates in index, the entry appearing first in ff will be deleted")
            eprint(" |   you can change this behaviour with [-z] or [-d]")

    if args.flatfile[-3:] == ".gz":
        eprint("    => ERROR: -f argument has extension .gz; wrong file??")
        sys.exit(22)

    if args.flatfile.find("://") != -1 and not args.remote:
        eprint("    => NOTICE: -f argument appears to be an URL; wrong file??")
        sys.exit(22)

    if args.outpath is None:
        args.outindex_filename = args.index_filename + ".new"
        args.output_filename = args.flatfile + ".new"
    else:
        if not os.access(args.outpath, os.W_OK):
            eprint("    => ERROR: specified outpath '{}' doesn't exist or is not writable!".format(
                args.outpath))
            sys.exit(1)
        args.outindex_filename = os.path.join(args.outpath, args.index_filename + ".new")
        args.output_filename = os.path.join(args.outpath, args.flatfile + ".new")
    if args.verbose:
        eprint(" |-- updated flatfile and index will be '{}' '{}'".format(
            args.output_filename,
            args.outindex_filename))

    #gather information from first line of index
    args.encrypted = False
    with open(args.index_filename, 'r', 1) as indexfh:
        args.index_type, _, _, _ = check_index(indexfh.readline())
    if args.index_type in (".", "+"):
        args.encrypted = True
        if args.verbose:
            eprint(" |-- index made for encrypted entries")
            eprint(" `=> Please ensure the -f filename points to the encrypted flatfile")
    if args.index_type in (":", "+"):
        if args.verbose:
            eprint(" |-- index made for compressed entries")
            eprint(" `=> Please ensure the -f filename points to the compressed flatfile")

    if args.index_blocksize is not None:
        if args.threads is None:
            eprint("    => ERROR: specifying blocksize makes sense only for -t execution")
            sys.exit(22)
        if args.verbose:
            eprint(" |-- blocksize set to {} bytes".format(args.index_blocksize))
    if args.threads is not None: #multithread
        if args.threads < 2:
            eprint("    => ERROR: No sense specifying a number of threads lower than 2!")
            sys.exit(22)
        if args.index_blocksize is None:
            #if not specified, we use 1/threadnumTH of filesize up to a minimum MINBLOCKSIZE
            args.index_blocksize = max(
                calculate_blocksize(args.index_filename, args.threads),
                siprefix2num(MINBLOCKSIZE))
        args.mt_subfiles_dir = TEMPDIR + "/tmpREINDEX" + randnum + "/"
        args.mt_subfiles_iprefix = args.mt_subfiles_dir + "I"
        args.mt_subfiles_oprefix = args.mt_subfiles_dir + "O"
        args.list_filesize, args.chunks_count = calculate_chunknum(
            args.index_filename, args.index_blocksize)
        if args.verbose:
            eprint(" |-- Parallel work in {} chunks of maximum {} bytes (-b to change)".format(
                args.chunks_count, args.index_blocksize))
            eprint(" |-- using maximum {} parallel threads (-t); your OS reports {} cpus.".format(
                args.threads, os.cpu_count()))
    else: # if unspecified, set args.threads to 1
        args.threads = 1
        args.chunks_count = 1


def check_files():
    """
    do some checks on availability of resources
    """
    check_iofiles([args.flatfile], [])
    check_iofiles([], [args.output_filename, args.outindex_filename])
    if args.threads > 1: #multithread
        os.mkdir(args.mt_subfiles_dir, mode=0o700)
        args.chunk_itempfiles = list()
        for chunknum in range(args.chunks_count):
            chunknumstr = str(chunknum).zfill(len(str(args.chunks_count))) #e.g. 00, 01..
            chunk_tempfile = args.mt_subfiles_iprefix + "." + chunknumstr
            try:
                myoutputfh = open(chunk_tempfile, 'w')
            except PermissionError:
                eprint("    => ERROR: Cannot open temporary file '{}' for writing".format(
                    chunk_tempfile))
                sys.exit(1)
            args.chunk_itempfiles.append(chunk_tempfile) #store temp filenames
            myoutputfh.close()
        delete_files(args.chunk_itempfiles)

        args.chunk_otempfiles = list()
        for chunknum in range(args.chunks_count):
            chunknumstr = str(chunknum).zfill(len(str(args.chunks_count))) #e.g. 00, 01..
            chunk_tempfile = args.mt_subfiles_oprefix + "." + chunknumstr
            try:
                myoutputfh = open(chunk_tempfile, 'w')
            except PermissionError:
                eprint("    => ERROR: Cannot open temporary file '{}' for writing".format(
                    chunk_tempfile))
                sys.exit(1)
            args.chunk_otempfiles.append(chunk_tempfile) #store temp filenames
            myoutputfh.close()
        delete_files(args.chunk_otempfiles)


def collect_entry(indexfh, identifier):
    """
    find position and size of entries corresponding to specified identifier
    from index
    """
    if args.duplicates:
        positions = get_positions(indexfh, identifier)
    elif args.zfound:
        positions = [get_position_last(indexfh, identifier)]
    else:
        positions = [get_position_first(indexfh, identifier)]

    if not positions or positions == [None]: #empty array
        if args.verbose:
            eprint("    => WARNING: '{}' not found in index; skipping".format(identifier))
        return [], []

    entry_positions = list()
    entry_lengths = list()

    for position in positions:
        #1) extract position and entry size
        if args.encrypted: #with iv
            posmatch = REESIVN.match(position)
        else: #no iv
            posmatch = REES.match(position)
        entry_position, entry_length = posmatch.groups()

        #2) decode and append to lists
        entry_positions.append(b64_to_int(entry_position))
        entry_lengths.append(b64_to_int(entry_length))

    return entry_positions, entry_lengths


def collect_entries_to_delete(list_filename, index_filename):
    """
    collect information about entries to delete, consulting the index,
    return a sorted list and a dict
    """
    indexfh = open(index_filename, 'r', 1)

    all_entries_positions = SortedList()
    all_entries_position2size = dict()

    identifiers_count = 0
    with open(index_filename) as indexfh:
        with open(list_filename) as listfh:
            for line in listfh:
                identifier = line.rstrip()
                identifiers_count += 1
                entry_positions, entry_sizes = collect_entry(indexfh, identifier)
                if entry_positions:
                    #eprint("got positions {} and sizes {} for identifier {}".format(
                    #   entry_positions, entry_sizes, identifier)) #debug
                    all_entries_positions.update(entry_positions) #add to sorted list
                    for entry_position, entry_size in zip(entry_positions, entry_sizes):
                        all_entries_position2size[entry_position] = entry_size #update dict

    return all_entries_positions, all_entries_position2size, identifiers_count


def delete_entries(inputfile, outputfile, sorted_positions, position2size):
    """
    delete the entries from input_filename writing into output_filename
    according to given positions and sizes
    """
    current_position = 0 #start from beginning
    with open(inputfile, 'rb') as inputfh:
        with open(outputfile, 'wb') as outfh:
            for entry_position in sorted_positions:
                entry_size = position2size[entry_position]
                write_size = entry_position - current_position
                if write_size > 0: #no sense writing nothing
                    inputfh.seek(current_position)
                    #eprint("  .. writing {} bytes from {}".format( #debug
                    #   write_size, current_position)) #debug
                    #buffered write:
                    blockcount, remainder = divmod(write_size, BUFFERSIZE)
                    for _ in range(blockcount):
                        buffered = inputfh.read(BUFFERSIZE)
                        outfh.write(buffered)
                    outfh.write(inputfh.read(remainder))
                current_position += write_size + entry_size #skip to after the entry being deleted

            #write the last piece of inputfile after last entry
            write_size = os.path.getsize(inputfile) - current_position
            if write_size > 0: #no sense writing nothing
                inputfh.seek(current_position)
                #eprint("  .. writing {} bytes from {}".format( #debug
                #   write_size, current_position)) #debug
                #buffered write:
                blockcount, remainder = divmod(write_size, BUFFERSIZE)
                for _ in range(blockcount):
                    buffered = inputfh.read(BUFFERSIZE)
                    outfh.write(buffered)
                outfh.write(inputfh.read(remainder))
            current_position += entry_size #skip to after the entry being deleted

            outfh.flush()


def find_shift_offset(position, sorted_positions, position2offset):
    """
    Find the offset to apply to a position,
    depending on cumulative sizes of entries deleted before it
    """
    first_deleted_position = sorted_positions[0]
    offset = 0
    if position < first_deleted_position:
        #quick heuristic: no need to check if position lower than first deleted
        #eprint("{}: before any deletions, no change".format(position)) #DEBUG
        return offset
    for deleted_position in reversed(sorted_positions):
        if position >= deleted_position:
            offset = position2offset[deleted_position]
            #eprint("{} >= {} so should decrement by {}".format( #debug
            #   position, deleted_position, offset)) #debug
            break
    #if offset == 0: #debug
    #   eprint("{}: no change".format(position)) #debug
    return offset


def update_index_after_deletions(index_filename, outfile,
                                 sorted_positions, position2offset,
                                 index_type):
    """
    update index file, creating outindex with entry positions shifted according to
    information given
    """
    #go through index and shift positions according to dict with offsets
    #for how much to shift entries found after a specified position
    buffered = ""
    all_indexes_count = 0
    skipped_count = 0
    shifted_count = 0
    unshifted_count = 0
    buflines = 100 #write every 100 lines
    bufcounter = 0
    with open(index_filename, 'r', 1) as indexfh: #using line buffering
        with open(outfile, 'w') as outfh:
            for line in indexfh:
                columns = line.rstrip().split(FIELDSEPARATOR)
                position = columns[1] #2nd column of index file is position
                if index_type in ("+", "."): #entry sizes and iv stored
                    posmatch = REESIVN.match(position)
                elif index_type in (":", "-"): #entry sizes stored, no iv
                    posmatch = REES.match(position)
                position, remaining = posmatch.groups()
                position = b64_to_int(position)
                if position in position2offset:
                    skipped_count += 1
                    continue
                if bufcounter > buflines: #write to outfile and reset buffer
                    outfh.write(buffered)
                    all_indexes_count += bufcounter
                    bufcounter = 0
                    buffered = ""

                offset = find_shift_offset(position, sorted_positions, position2offset)
                if offset > 0:
                    columns[1] = "{}{}{}".format(int_to_b64(position - offset),
                                                 index_type,
                                                 remaining)
                else:
                    unshifted_count += 1
                bufcounter += 1
                buffered += FIELDSEPARATOR.join(columns) + "\n"
            all_indexes_count += bufcounter
            outfh.write(buffered) #last block
            shifted_count = all_indexes_count - unshifted_count
            all_indexes_count += skipped_count
    return all_indexes_count, shifted_count, skipped_count


def print_stats(start_time):
    """
    print some final statistics on entries deleted
    """
    indexes_sum = 0
    deleted_sum = 0
    reindexed_sum = 0
    for chunknum in range(args.chunks_count):
        indexes_sum += indexes_count[chunknum]
        reindexed_sum += reindexed_count[chunknum]
        deleted_sum += deleted_count[chunknum]

    if found_count == 1:
        eprint(" |-- Found and removed 1 entry.")
    elif found_count > 0:
        eprint(" |-- Found and removed {} entries.".format(
            found_count))
    if found_count < requested_count:
        if found_count == 0:
            eprint("    => WARNING: NONE of the {} requested identifiers found in index!".format(
                requested_count))
        else:
            eprint("    => WARNING: only {} of the {} requested identifiers found in index.".format(
                found_count, requested_count))
    eprint(" |-- Deleted {} and reindexed {} indexes out of total {}.".format(
        deleted_sum, reindexed_sum, indexes_sum))
    _, reindexed_speed = elapsed_time(start_time, reindexed_sum)
    eprint(" '-- Elapsed: {}, {} deletions/sec; {} reindexing/sec --'".format(
        *elapsed_time(start_time, found_count), reindexed_speed))


def update_index_wrapper(chunknum):
    """
    wrapper to shift index for multithread use
    """
    #global reindexed_count, deleted_count
    if args.threads > 1:
        index_filename = args.chunk_itempfiles[chunknum]
        outindex_filename = args.chunk_otempfiles[chunknum]
    else:
        index_filename = args.index_filename
        outindex_filename = args.outindex_filename

    indexes_count[chunknum], reindexed_count[chunknum], \
        deleted_count[chunknum] = update_index_after_deletions(
            index_filename, outindex_filename, mysorted_positions,
            myposition2offset, args.index_type)


def init_thread(i, r, d):
    """
    to initialise multithreaded worker
    """
    global indexes_count, reindexed_count, deleted_count
    indexes_count = i
    reindexed_count = r
    deleted_count = d


if __name__ == '__main__':
    check_args()
    check_files()

    if args.output_filename is None:
        args.outputfh = sys.stdout
    else:
        args.outputfh = open(args.output_filename, 'wb')

    start_secs = time.time()

    #1) collect information from index on entries to delete
    mysorted_positions, myposition2size, \
        requested_count = collect_entries_to_delete(args.list_filename, args.index_filename)
    found_count = len(mysorted_positions)

    #2) delete entries from flatfile
    delete_entries(args.flatfile, args.output_filename, mysorted_positions, myposition2size)

    #3) fill dict with cumulative offsets, used to update index
    myposition2offset = dict()
    size_offset = 0
    for myentry_position in mysorted_positions:
        size_offset += myposition2size[myentry_position]
        myposition2offset[myentry_position] = size_offset
    #eprint("removed a total of {} bytes".format(size_offset)) #debug
    if os.path.getsize(args.flatfile) - size_offset != os.path.getsize(args.output_filename):
        eprint("    => ERROR: problems with deletion, file size of resulting file is wrong")
        sys.exit(1)

    #4) update the index shifting positions, optionally with multithread
    if args.threads > 1: #multithread
        args.chunk_itempfiles, _ = split_file(args.index_filename,
                                              args.index_blocksize,
                                              args.mt_subfiles_iprefix)
        args.chunks_count = len(args.chunk_itempfiles)

        if args.verbose:
            eprint(" |-- parallel reindexing in chunks of maximum {} bytes (-b to change)".format(
                args.index_blocksize))
            eprint(" |-- index will be split into {} chunks".format(args.chunks_count))

        indexes_count = Array('i', [0] * args.chunks_count)
        reindexed_count = Array('i', [0] * args.chunks_count)
        deleted_count = Array('i', [0] * args.chunks_count)

        args.chunk_otempfiles = args.chunk_otempfiles[0:args.chunks_count]

        #init threads
        pool = Pool(args.threads, initializer=init_thread,
                    initargs=(indexes_count, reindexed_count, deleted_count))
        #submit chunks to threads
        if args.progressbar:
            _ = list(tqdm(pool.imap(update_index_wrapper, range(args.chunks_count)),
                          total=args.chunks_count, ascii=PROGRESSBARCHARS))
        else:
            pool.imap(update_index_wrapper, range(args.chunks_count))
        pool.close() #no more work to submit
        pool.join() #wait workers to finish

        delete_files(args.chunk_itempfiles) #cleanup

        #final union of results: combine shifted index chunks
        with open(args.outindex_filename, 'wb') as outputfh:
            print_subfiles(args.chunk_otempfiles, outputfh)

        delete_files(args.chunk_otempfiles) #cleanup
    else:
        indexes_count = [0]
        deleted_count = [0]
        reindexed_count = [0]
        update_index_wrapper(0)

    print_stats(start_secs)



import setuptools

with open("README_brief.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ffdb",
    version="2.4.0",
    author="Dr Giuseppe Insana",
    author_email="insana@insana.net",
    description="ffdb",
    long_description=long_description,
    long_description_content_type="text/markdown",
    scripts=["scripts/indexer.py", "scripts/extractor.py",
             "scripts/remover.py", "scripts/merger.py"],
    url="https://github.com/g-insana/ffdb.py",
    license="AGPL",
    py_modules=["ffdb"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "sortedcontainers",
        "requests",
        "pycryptodomex"
    ],
    python_requires=">=3.5",
)



#!/usr/bin/env python3
"""
merger.py
objective: merge new entries to existing flatfile, updating its index

by Giuseppe Insana, 2020
"""

# pylint: disable=C0103,R0912,R0915,W0603

#IMPORTS:
import os
import sys
import argparse
from random import randint
from shutil import copyfile
from ffdb import eprint, check_index, delete_files, GZTOOL_EXE, \
    merge_indexes, TEMPDIR, BUFFERSIZE

#CONSTANTS
PROGNAME = "merger.py"
VERSION = "1.5"
AUTHOR = "Giuseppe Insana"
args = None


def check_index_compatibility(indexfile1, indexfile2):
    """
    will check if old and new indexes are of same type
    """

    with open(indexfile1, 'r', 1) as indexfh:
        line1 = indexfh.readline() #first line of index

    index_type1, cipher1, keysize1, has_checksum1 = check_index(line1)

    with open(indexfile2, 'r', 1) as indexfh:
        line2 = indexfh.readline() #first line of index

    index_type2, cipher2, keysize2, has_checksum2 = check_index(line2)

    if index_type1 != index_type2 or cipher1 != cipher2 or keysize1 != keysize2:
        eprint("    => ERROR: The indexes you are merging '{}' and '{}'".format(
            indexfile1, indexfile2))
        eprint("              are of incompatible type! Cannot proceed!")
        sys.exit(22)
    if has_checksum1 != has_checksum2:
        eprint("    => ERROR: The indexes you are merging '{}' and '{}'".format(
            indexfile1, indexfile2))
        eprint("              are of the same type but only one contains checksums!")
        eprint("       Both should be either with or without checksums. Cannot proceed!")
        sys.exit(22)
    args.index_type = index_type1


def uncompress_file(filename):
    """
    uncompress given filename
    return name of the uncompressed file
    """
    return filename


def compress_file(filename, create_gzindex=None):
    """
    compress given filename
    optionally create gzindex
    return name of the uncompressed file
    """
    if create_gzindex is not None:
        #open pipe to call external gztool
        pass
    return filename


def concatenate_into(filename, addenda, outfilename):
    """
    first copy original file to a new file
    then append content of addenda file at the end of the new file
    """
    copyfile(filename, outfilename)
    #append_to(outfilename, addenda) #reads the whole file in memory
    append_to_buffered(outfilename, addenda)


def append_to_buffered(filename, addenda):
    """
    append addenda file at the end of the given file, using buffered read/write
    """
    with open(addenda, 'rb') as inputfh:
        with open(filename, 'ab') as outfh:
            ##straight, unbuffered, will read whole file in memory:
            #outfh.write(inputfh.read())
            ##buffered:
            blockcount, remainder = divmod(os.path.getsize(addenda), BUFFERSIZE)
            for _ in range(blockcount):
                buffered = inputfh.read(BUFFERSIZE)
                outfh.write(buffered)
            outfh.write(inputfh.read(remainder))
            outfh.flush()


def check_args():
    """
    parse arguments and check for error conditions
    """
    global args, GZTOOL_EXE
    usagetxt = """{0} -f FLATFILE -i INDEXFILE -e ENTRIESFILE -n NEWINDEXFILE
    [-f] : flatfile into which the new entries should be added
    [-i] : index of FLATFILE
    [-e] : filename containing the new entries to be added
    [-n] : index of ENTRIESFILE
    see {0} -h for tweaks and optional modes
    \nexamples:
       {0} -f db.dat -i db.pos -e new.dat -n new.pos
         (will update db.dat and db.pos)
       {0} -c -f db.dat -i db.pos -e new.dat -n new.pos
         (will create db.dat.new and db.pos.new)
       {0} -c -o export -f db.dat -i db.pos -e new.dat -n new.pos
         (will create export/db.dat.new and export/db.pos.new)
    """.format(PROGNAME)
    parser = argparse.ArgumentParser(description='Merge new pre-indexed entries into an existing \
                                     flatfile', usage=usagetxt)
    parser.add_argument('-f', '--file', dest='ff_filename',
                        help="filename of flatfile to be processed",
                        required=True, type=str)
    parser.add_argument('-i', '--index', dest='index_filename',
                        help="filename of index file with entry identifiers",
                        required=True, type=str)
    parser.add_argument('-e', '--entries', dest='newentries_filename',
                        help="filename of new entries to be merged into flatfile",
                        required=True, type=str)
    parser.add_argument('-n', '--newindex', dest='newindex_filename',
                        help="filename of index file with entry identifiers",
                        required=True, type=str)
    parser.add_argument('-c', '--create', dest='createmode', action='store_true',
                        help="create new files (.new extension) rather than updating existing \
                        files (the default operation mode)", required=False)
    parser.add_argument('-o', '--outpath', dest='outpath',
                        help="optionally write new files to specified path rather than creating \
                        new files in the same location as the original ones",
                        required=False, type=str)
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help="verbose operation", required=False)
    parser.add_argument('-d', '--delete', dest='deleteafter', action='store_true',
                        help="delete ENTRIESFILE and NEWINDEXFILE after merging is completed",
                        required=False)
    parser.add_argument('-g', '--gzip', dest='gzip', action='store_true',
                        help="compress the final flatfile after merge, creating .gzi compressed \
                        index", required=False)
    parser.add_argument('-s', '--small', dest='smallnew', action='store_true',
                        help="use this mode if the new index is small (<30k entries): performance \
                        should be better", required=False)
    args = parser.parse_args()

    if args.verbose:
        eprint(" .-- {} v{} -- by {} --.".format(PROGNAME, VERSION, AUTHOR))

    args.ff_compressed = False
    if args.ff_filename[-3:] == ".gz":
        args.ff_compressed = True
        if not args.gzip:
            eprint("    => NOTICE: -f argument has extension .gz: assuming flatfile is compressed")
            eprint("               it will be uncompressed and then recompressed after merge")
            args.gzip = True

    args.newentries_compressed = False
    if args.newentries_filename[-3:] == ".gz":
        args.newentries_compressed = True
        eprint("    => NOTICE: -n argument has extension .gz: assuming newentriesfile compressed")

    if args.ff_filename.find("://") != -1:
        eprint("    => ERROR: {} cannot operate on remote flatfiles".format(PROGNAME))
        sys.exit(22)

    if args.smallnew:
        if args.verbose:
            eprint(" |-- using tweak for smaller new index")
    else:
        randnum = str(randint(1000, 9999))
        args.itemp_filename = TEMPDIR + "/tmpMERGE" + randnum

    if args.outpath is None:
        args.outindex_filename = args.index_filename + ".new"
    else:
        if not os.access(args.outpath, os.W_OK):
            eprint("    => ERROR: specified outpath '{}' doesn't exist or is not writable!".format(
                args.outpath
            ))
            sys.exit(1)
        args.outindex_filename = os.path.join(args.outpath, args.index_filename + ".new")


def check_iofiles(read_filenames, write_filenames):
    """
    check for ability to open input/output filenames
    """
    if read_filenames is not None:
        for filename in read_filenames:
            try:
                inputfh = open(filename, 'r')
                inputfh.close()
            except FileNotFoundError:
                eprint("    => ERROR: Cannot open file '{}' for reading".format(filename))
                sys.exit(2)

    if write_filenames is not None:
        for filename in write_filenames:
            if os.path.isfile(filename):
                eprint("    => ERROR: file '{}' exists".format(filename))
                eprint("       please remove it as we refuse to overwrite it!")
                sys.exit(1)
            try:
                myoutputfh = open(filename, 'w')
                myoutputfh.close()
            except PermissionError:
                eprint("    => ERROR: Cannot open file '{}' for writing".format(filename))
                sys.exit(1)
            #eprint("deleting {}".format(filename)) #DEBUG
            delete_files([filename])


def check_files():
    """
    check for ability to open input/output filenames
    """
    check_iofiles([args.ff_filename, args.newentries_filename, args.index_filename,
                   args.newindex_filename], [args.outindex_filename])

    if not args.smallnew:
        check_iofiles(None, [args.itemp_filename])

    if args.createmode:
        if args.outpath is None:
            args.outff_filename = args.ff_filename + ".new"
        else:
            args.outff_filename = os.path.join(args.outpath, args.ff_filename + ".new")
        check_iofiles(None, [args.outff_filename])


if __name__ == '__main__':
    #parse and check arguments
    check_args()

    #check files (if they can be read/written)
    check_files()

    #check if old and new indexes are of same type:
    check_index_compatibility(args.index_filename, args.newindex_filename)

    #uncompress files if needed
    if args.ff_compressed:
        if args.verbose:
            eprint(" |-- uncompressing original flatfile.. this may take some time..")
        args.ff_filename = uncompress_file(args.ff_filename)
    if args.newentries_compressed:
        if args.verbose:
            eprint(" |-- uncompressing newentries file.. this may take some time..")
        args.newentries_filename = uncompress_file(args.newentries_filename)

    #calculate index offset
    pos_offset = os.path.getsize(args.ff_filename)

    #merge old and new identifiers' indexes
    if args.smallnew:
        tempfile = None
    else:
        tempfile = args.itemp_filename
    new_identifiers_count = merge_indexes(args.index_filename, args.newindex_filename,
                                          pos_offset, args.index_type,
                                          args.outindex_filename, tempfile)

    #create new flatfile or update original flatfile
    if args.createmode:
        if args.verbose:
            eprint(" |-- new indexfile created: '{}'".format(args.outindex_filename))
        concatenate_into(args.ff_filename, args.newentries_filename, args.outff_filename)
        if args.gzip: #compress flatfile
            if args.verbose:
                eprint(" |-- compressing new flatfile.. this may take some time..")
            args.outff_filename = compress_file(args.outff_filename, create_gzindex=True)
            if args.verbose:
                eprint(" |-- new compressed flatfile created: '{}'".format(args.outff_filename))
        else:
            if args.verbose:
                eprint(" |-- new flatfile created: '{}'".format(args.outff_filename))
    else: #updatemode
        os.replace(args.outindex_filename, args.index_filename)
        if args.verbose:
            eprint(" '-- original indexfile updated: '{}'".format(args.index_filename))
        append_to_buffered(args.ff_filename, args.newentries_filename)
        if args.gzip: #compress flatfile
            if args.verbose:
                eprint(" |-- compressing updated flatfile.. this may take some time..")
            args.ff_filename = compress_file(args.ff_filename, create_gzindex=True)
            if args.verbose:
                eprint(" |-- original flatfile updated and compressed: '{}'".format(
                    args.ff_filename))
        else:
            if args.verbose:
                eprint(" |-- original flatfile updated: '{}'".format(args.ff_filename))

    #final cleanup, if desired
    if args.deleteafter:
        delete_files([args.newentries_filename, args.newindex_filename])

    #final printout
    eprint(" '-- merged {} new entries".format(new_identifiers_count))



#!/usr/bin/env python3
"""
ffdb
This module and the associated set of utility scripts
allow the creation, maintenance and usage of a database and document storage
which employs a single container file.

by Giuseppe Insana, 2020
"""
import os
import re
import sys
import time
import zlib
from math import ceil
from heapq import merge #alternative mergesort
from Cryptodome.Cipher import AES
from Cryptodome.Protocol import KDF
from Cryptodome import Random

# pylint: disable=C0103,R0912,R0915,W0603
VERSION = '2.4.0'

##CUSTOMIZATIONS:

#personalize this program by modifying the
#following SALT used by the PBKDF2 routine
SALT = b'5ed3a4284d6a9c1e4e4f6b4729b254be'

#change the following if you prefer space or other delimiter for the index
FIELDSEPARATOR = "\t"

#define temporary directory in case no TMPDIR is specified
if 'TMPDIR' in os.environ and os.environ['TMPDIR'] != '':
    TEMPDIR = os.environ['TMPDIR']
else:
    TEMPDIR = "/tmp"

#customize the progress bar (for multithread verbose progress)
#PROGRESSBARCHARS = False #use this for standard unicode smooth blocks
PROGRESSBARCHARS = ' ·•'

#only relevant for extraction via gztool
#from local or remote compressed flatfiles;
#what is the full path of gztool:
GZTOOL_EXE = None
#e.g.:
#GZTOOL_EXE = "/usr/local/bin/gztool"
#if not hardcoded here, be sure it is found in your PATH

#buffer size used by io operations
#BUFFERSIZE = 8192
BUFFERSIZE = 1048576

##CONSTANTS:

REIV = re.compile(r'^([0-9a-zA-Z{}]+)\.([0-9a-zA-Z{}]+)\|([A-Z])(.+)$') #only encrypted
REIVD = re.compile(r'^([0-9a-zA-Z{}]+)\+([0-9a-zA-Z{}]+)\|([A-Z])(.+)$') #deflated and encrypted
REESIV = re.compile(r'^([0-9a-zA-Z{}]+)(?:\.|\+)([0-9a-zA-Z{}]+)\|[A-Z](.+)$') #entrysize+iv stored
REESIVN = re.compile(r'^([0-9a-zA-Z{}]+)(?:\.|\+)([0-9a-zA-Z{}]+\|[A-Z].+)$') #don't capture iv
REES = re.compile(r'^([0-9a-zA-Z{}]+)(?::|-)([0-9a-zA-Z{}]+)$') #entrysize but no iv stored
RETERMINATOR = re.compile(r'^-$')

B64_CHARS = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ{}'
B64_INDEX = dict((char, pos) for pos, char in enumerate(tuple(B64_CHARS)))


def calculate_chunknum(filename, blocksize):
    """
    calculate number of chunks from global defined blocksize
    and file's size
    """
    filesize = os.path.getsize(filename)
    number_of_chunks = filesize // blocksize + 1
    #eprint("filename {} filesize {} blocksize {}".format(filename, filesize, blocksize)) #debug
    return filesize, number_of_chunks


def calculate_blocksize(filename, number_of_chunks):
    """
    calculate size of a single block
    from number of chunk/splits desired and a file's filesize
    """
    filesize = os.path.getsize(filename)
    if filesize == 0:
        raise RuntimeError("File '{}' is empty".format(filename))
    blocksize = int(ceil(filesize / number_of_chunks))
    number_of_chunks = int(ceil(filesize / blocksize))
    return blocksize


def find_next_delimiter(fp, filesize, startpos, delimiter, buffersize=None):
    """
    return the position of first delimiter found after given starting position
    """
    if startpos > filesize:
        return filesize
    if buffersize is None:
        buffersize = 16384
    delimiterlen = len(delimiter)
    fp.seek(startpos)
    data = fp.read(buffersize)
    while data:
        delimiterpos = data.find(delimiter)
        #eprint("{}-{} checking for {}: {}".format(startpos, fp.tell(), delimiter, delimiterpos))
        if delimiterpos != -1:
            return startpos + delimiterpos + delimiterlen
        startpos += buffersize
        data = fp.read(buffersize)
    return filesize #could not find


def find_next_newline(fp, filesize, startpos):
    """
    return the position of first newline found after given starting position
    """
    if startpos > filesize:
        return filesize
    if startpos <= 0:
        return 0
    fp.seek(startpos - 1)
    fp.readline() # to avoid truncatedlines
    return fp.tell()


def chunk_startpos_onnewline(fp, filesize, blocksize, block):
    """
    return the initial byte of a chunk without breaking newlines
    """
    ini = block * blocksize
    end = ini + blocksize
    if ini > filesize:
        return None
    if end > filesize:
        end = filesize
    if ini <= 0:
        fp.seek(0)
    else:
        fp.seek(ini - 1)
        fp.readline() # to avoid truncatedlines
    return fp.tell()


def chunk_of_lines(fp, filesize, blocksize, block):
    """
    return a generator of the lines coming from a single chunk of the file
    """
    ini = block * blocksize
    end = ini + blocksize
    if ini > filesize:
        return
    if end > filesize:
        end = filesize
    if ini <= 0:
        fp.seek(0)
    else:
        fp.seek(ini - 1)
        fp.readline() # to avoid truncatedlines
    while fp.tell() < end:
        yield fp.readline()


def int_to_b64(number):
    """
    convert a decimal integer to a base64encoded number
    """
    if number == 0:
        return '0'
    string = ''
    while number:
        string = B64_CHARS[number % 64] + string
        number //= 64
        #~13% slower:
        #number, remainder = divmod(number, 64)
        #string = B64_CHARS[remainder] + string
    return string


def b64_to_int(string):
    """
    convert a base64encoded number to decimal integer
    """
    strlen = len(string)
    number = 0
    index = 0
    for char in string:
        index += 1
        power = strlen - index
        number += B64_INDEX[char] * (64 ** power)
        #~20% slower:
        # number += B64_CHARS.index(char) * (64 ** power)
    return number


def format_indexes(entry, index_type, cipher_type, checksums=False):
    """
    return array of formatted indexes
    """
    fieldseparator = FIELDSEPARATOR
    indexes = list()
    #if len(entry['ids']) == 0:
    #   eprint("processing entry '{}'".format(entry['full']))
    if index_type == "":
        for identifier in entry['ids']:
            indexes.append("{}\n".format(identifier))
        return indexes

    length = entry['length_b64']
    if index_type in (".", "+"): #encryption: the iv is part of the index
        entryposition = "{}{}{}|{}{}".format(entry['position'],
                                             index_type,
                                             length,
                                             cipher_type,
                                             entry['iv'].hex())
    else: #no iv
        entryposition = "{}{}{}".format(entry['position'],
                                        index_type,
                                        length)
    for identifier in entry['ids']:
        if checksums:
            indexes.append("{1}{0}{2}{0}{3}\n".format(fieldseparator,
                                                      identifier,
                                                      entryposition,
                                                      int_to_b64(entry['checksum'])))
        else:
            indexes.append("{}{}{}\n".format(identifier, fieldseparator, entryposition))
    return indexes


def deflate(data, compresslevel=9):
    """
    compress a byte object
    """
    compress = zlib.compressobj(
        compresslevel,        # level: 0-9
        zlib.DEFLATED,        # method: must be DEFLATED
        -zlib.MAX_WBITS,      # window size in bits:
                              #   -15..-8: negate, suppress header
                              #   8..15: normal
                              #   16..30: subtract 16, gzip header
        zlib.DEF_MEM_LEVEL,   # mem level: 1..8/9
        0                     # strategy:
                              #   0 = Z_DEFAULT_STRATEGY
                              #   1 = Z_FILTERED
                              #   2 = Z_HUFFMAN_ONLY
                              #   3 = Z_RLE
                              #   4 = Z_FIXED
    )
    deflated = compress.compress(data)
    deflated += compress.flush()
    return deflated


def inflate(data):
    """
    decompress a byte object
    """
    decompress = zlib.decompressobj(
        -zlib.MAX_WBITS
    )
    inflated = decompress.decompress(data)
    inflated += decompress.flush()
    return inflated


def get_cipher_type(cipher):
    """
    return cipher_type letter from cipher
    """
    cipher2letter = {'aes128': 'A', 'aes192': 'B', 'aes256': 'C'}
    if cipher not in cipher2letter:
        eprint("    => ERROR: cipher '{}' not recognised!".format(cipher))
        sys.exit(1)
    return cipher2letter[cipher]


def find_keysize_cipher(cipher_type):
    """
    infer cipher and keysize from cipher_type letter
    return keysize and cipher
    """
    if cipher_type == "A":
        cipher_name = "aes128"
        keysize = 16
    elif cipher_type == "B":
        cipher_name = "aes192"
        keysize = 24
    elif cipher_type == "C":
        cipher_name = "aes256"
        keysize = 32
    else:
        eprint("    => ERROR: cipher_type '{}' not recognised!".format(cipher_type))
        sys.exit(1)
    return keysize, cipher_name


def merge_indexes(index1_filename, index2_filename, pos_offset,
                  index_type, outfile, tempfile=None):
    """
    merge two index files into one outfile, shifting the second one of given offset
    returns number of identifiers from the second index
    """
    if tempfile is None: #read second index into array, then merge
        new_identifiers = shift_index(index2_filename,
                                      pos_offset,
                                      index_type)
        new_identifiers_count = len(new_identifiers)
        #faster than heap_mergesort if len(new_identifiers) is small
        mergesort_file_array(index1_filename, new_identifiers, outfile)
        #slower but with good performance no matter which size:
        #heap_mergesort_file_array(index1_filename, new_identifiers, outfile) #ALT
    else: #use a temporary file to shift the second index, then merge files
        new_identifiers_count = shift_index_file(index2_filename,
                                                 pos_offset,
                                                 tempfile,
                                                 index_type)
        mergesort_files([index1_filename, tempfile], outfile)
        os.remove(tempfile)
    return new_identifiers_count


def mergesort_file_array(filename, insertions, outfilename):
    """
    merge sort contents of a file and an array
    writing to outputfile
    """
    newline = insertions.pop(0)
    with open(outfilename, 'w') as outfh:
        with open(filename, 'r', 1) as sortedfh: #using line buffering
            for line in sortedfh:
                while insertions:
                    if newline <= line:
                        outfh.write(newline)
                        newline = insertions.pop(0) #prepare for next
                    else:
                        break
                if newline <= line:
                    outfh.write(newline)
                    newline = ""
                outfh.write(line)
            if newline != "":
                outfh.write(newline)
            if insertions: #if some left in the end, append
                for newline in insertions:
                    outfh.write(newline)


def heap_mergesort_file_array(filename, insertions, outfilename):
    """
    use heapq.merge to merge sort contents of a file and an array
    writing to outputfile
    """
    with open(outfilename, 'w') as outfh:
        with open(filename, 'r', 1) as sortedfh: #using line buffering
            outfh.writelines(merge(sortedfh, insertions))


def mergesort_files(filenames, outfilename):
    """
    use heapq.merge to merge sort contents of files
    writing to outputfile
    """
    filehandles = list()
    for filename in filenames:
        filehandles.append(open(filename, 'r', 1))
    with open(outfilename, 'w') as outfh:
        outfh.writelines(merge(*filehandles))
    for filehandle in filehandles:
        filehandle.close()


def file_string_binary_search(fh, string, mode=None):
    """
    perform a binary_search for a string in a sorted text file
    return all matches, unless mode specified (can be 'first' or 'last')
    """
    found = None
    foundpos = None
    lo = 0
    fh.seek(0, 2)
    hi = fh.tell() - 1 #filesize
    while lo < hi:
        mid = (hi + lo) >> 1
        if mid > 0:
            fh.seek(mid - 1)
            fh.readline() #to avoid truncated line we skip this one
            midf = fh.tell()
        else:
            midf = 0
            fh.seek(midf)
        line = fh.readline() #read next line
        #eprint("lo {} hi {} mid {} midf {}".format(lo, hi, mid, midf)) #debug
        if not line or string <= line:
            hi = mid
        else:
            lo = mid + 1
    if mid == lo: #range fully scanned
        foundpos = midf
    elif lo <= 0: #very beginning
        foundpos = 0
    if foundpos is None:
        line = fh.readline()
        #foundpos = fh.tell()
    #eprint("got fist match at pos {}".format(foundpos))
    if mode == 'first': # return earliest match
        if line.startswith(string):
            found = line
    elif mode == 'last': # return latest match
        while line.startswith(string): #continue skipping lines while they match
            found = line
            line = fh.readline()
    else: #multi: return all matches
        found = list()
        while line.startswith(string): #continue adding lines while they match
            found.append(line)
            line = fh.readline()
    return found


def _file_string_binary_search_any(fh, string):
    """
    perform a binary_search for a string in a sorted text file
    it will quickly return any match, not necessarily the first or the last in the file
    """
    found = None
    #start checking the very first line:
    fh.seek(0)
    line = fh.readline()
    if line.startswith(string):
        found = line
    else:
        lo = 0 #beginning of file
        fh.seek(0, 2)
        hi = fh.tell() #end of file
        while hi - lo > 1:
            mid = (hi + lo) // 2
            fh.seek(mid, 0)
            fh.readline() #to avoid truncated line we skip this one
            line = fh.readline() #read next line
            if line.startswith(string):
                found = line
                break
            if string > line:
                lo = mid
            else:
                hi = mid
        if found is None:
            fh.seek(hi)
            line = fh.readline()
            if line.startswith(string):
                found = line

    if found is not None:
        return found
    return None


def get_position_first(indexfh, identifier):
    """
    find from the index the position of the entry
    corresponding to specified identifier
    """
    identifier += FIELDSEPARATOR #e.g. "ABCDE\t"
    line = file_string_binary_search(indexfh, identifier, mode='first')
    if line:
        columns = line.rstrip().split(FIELDSEPARATOR)
        return columns[1]
    return None #if could not be found


def get_position_last(indexfh, identifier):
    """
    find from the index the position of the entry
    corresponding to specified identifier
    """
    identifier += FIELDSEPARATOR #e.g. "ABCDE\t"
    line = file_string_binary_search(indexfh, identifier, mode='last')
    if line:
        columns = line.rstrip().split(FIELDSEPARATOR)
        return columns[1]
    return None #if could not be found


def get_positions(indexfh, identifier):
    """
    find from the index all the positions
    corresponding to (possibly duplicated) specified identifier
    """
    positions = list()
    identifier += FIELDSEPARATOR #e.g. "ABCDE "
    lines = file_string_binary_search(indexfh, identifier)
    if lines:
        for line in lines:
            columns = line.rstrip().split(FIELDSEPARATOR)
            positions.append(columns[1]) #2nd column of index file is position
    return positions


def get_position_checksum_first(indexfh, identifier):
    """
    find from the index the position of the entry
    corresponding to specified identifier
    additionally return checksum
    """
    identifier += FIELDSEPARATOR #e.g. "ABCDE\t"
    line = file_string_binary_search(indexfh, identifier, mode='first')
    if line:
        columns = line.rstrip().split(FIELDSEPARATOR)
        return columns[1], columns[2]
    return None #if could not be found


def get_position_checksum_last(indexfh, identifier):
    """
    find from the index the position of the entry
    corresponding to specified identifier
    additionally return checksum
    """
    identifier += FIELDSEPARATOR #e.g. "ABCDE\t"
    line = file_string_binary_search(indexfh, identifier, mode='last')
    if line:
        columns = line.rstrip().split(FIELDSEPARATOR)
        return columns[1], columns[2]
    return None #if could not be found


def get_positions_checksums(indexfh, identifier):
    """
    find from the index all the positions
    corresponding to (possibly duplicated) specified identifier
    """
    positions = list()
    checksums = list()
    identifier += FIELDSEPARATOR #e.g. "ABCDE "
    lines = file_string_binary_search(indexfh, identifier)
    if lines:
        for line in lines:
            columns = line.rstrip().split(FIELDSEPARATOR)
            positions.append(columns[1]) #2nd column of index file is position
            checksums.append(columns[2]) #3rd column of index file is checksum
    return positions, checksums


def init_cipher(key, iv):
    """
    initialize and return a new cipher
    """
    return AES.new(key, AES.MODE_CFB, iv)


def generate_iv():
    """
    generate and return a new iv
    """
    return Random.new().read(AES.block_size)


def derive_key(passphrase, keysize):
    """
    derive key of desired bytesize from passphrase
    """
    key = KDF.PBKDF2(passphrase, SALT, keysize)
    if len(key) != keysize:
        msg = "Attention! Something has gone wrong when deriving key from passphrase!"
        raise RuntimeError(msg)
    ks2cipher = {16: 'aes128', 24: 'aes192', 32: 'aes256'}
    cipher_name = ks2cipher[keysize]
    return cipher_name, key


def check_index(line):
    """
    read first record of a positional index and gather needed information
    return index_type, cipher_name, keysize, has_checksums
    """
    cipher_name = None
    keysize = None
    has_checksums = False

    if line.find(FIELDSEPARATOR) == -1:
        eprint("    => ERROR: index does not have positional column. Wrong file??")
        sys.exit(1)

    columns = line.split(FIELDSEPARATOR)
    if len(columns) > 2: #3rd column is crc
        has_checksums = True
    position = columns[1] #2nd column of index file
    iv = None
    if position.find("+") != -1: #both compressed and encrypted
        index_type = "+"
        iv = REIVD.match(position)
    elif position.find(".") != -1: #encrypted but not compressed
        index_type = "."
        iv = REIV.match(position)
    elif position.find(":") != -1: #compressed but not encrypted
        index_type = ":"
    elif position.find("-") != -1: #not encrypted nor compressed but entry sizes stored
        index_type = "-"
    else:
        eprint("    => ERROR: index does not have index_type identifier. Wrong file??")
        sys.exit(1)

    if index_type in (".", "+"): #encrypted entries
        if iv is None:
            eprint("    => ERROR: missing iv, cannot infer cipher type and keysize!")
            sys.exit(1)
        cipher_type = iv.group(3) #A, B, C..
        keysize, cipher_name = find_keysize_cipher(cipher_type)
    return index_type, cipher_name, keysize, has_checksums


def read_from_size(fh, begin, size):
    """
    reads from filehandle a specified amount of bytes
    """
    fh.seek(begin)
    return fh.read(size)


def delete_files(filenames):
    """
    delete the temporary sub-files
    """
    for filename in filenames:
        if os.path.isfile(filename):
            os.remove(filename)


def close_subfiles(fhlist):
    """
    close the filehandles for the temporary sub-files
    """
    for fh in fhlist:
        fh.close()


def print_subfiles(filenames, fh=None):
    """
    read and print a series of temporary sub-files writing them sequentially
    to a given output filehandle or print them to stdout
    """
    mode = 'rb'
    if fh is None:
        fh = sys.stdout #print to stdout if no fh given
        mode = 'r'
    for filename in filenames:
        with open(filename, mode) as infile:
            ##straight, unbuffered, will read whole file in memory
            #fh.write(infile.read())
            ##buffered:
            filesize = os.path.getsize(filename)
            blockcount, remainder = divmod(filesize, BUFFERSIZE)
            for _ in range(blockcount):
                buffered = infile.read(BUFFERSIZE)
                fh.write(buffered)
            fh.write(infile.read(remainder))


def shift_index(index_filename, offset, index_type):
    """
    shift positions in index file by given offset
    return list with identifiers/positions
    """
    identifiers = list()
    with open(index_filename, 'r', 1) as indexfh: #using line buffering
        for line in indexfh:
            columns = line.split(FIELDSEPARATOR)
            position = columns[1] #2nd column of index file is position
            if index_type in ("+", "."): #entry sizes and iv stored
                posmatch = REESIVN.match(position)
            elif index_type in (":", "-"): #entry sizes stored, no iv
                posmatch = REES.match(position)
            position, remaining = posmatch.groups()
            position = b64_to_int(position)
            columns[1] = "{}{}{}".format(int_to_b64(position + offset),
                                         index_type,
                                         remaining)
            identifiers.append(FIELDSEPARATOR.join(columns) + "\n")
    return identifiers


def shift_index_file(index_filename, offset, outfile, index_type):
    """
    shift positions in index file by given offset creating outfile
    """
    buffered = ""
    shifted_count = 0
    buflines = 100 #write every 100 lines
    bufcounter = 0
    with open(index_filename, 'r', 1) as indexfh: #using line buffering
        with open(outfile, 'w') as outfh:
            for line in indexfh:
                columns = line.split(FIELDSEPARATOR)
                position = columns[1] #2nd column of index file is position
                if index_type in ("+", "."): #entry sizes and iv stored
                    posmatch = REESIVN.match(position)
                elif index_type in (":", "-"): #entry sizes stored, no iv
                    posmatch = REES.match(position)
                position, remaining = posmatch.groups()
                position = b64_to_int(position)
                if bufcounter > buflines: #write to outfile and reset buffer
                    outfh.write(buffered)
                    shifted_count += bufcounter
                    bufcounter = 0
                    buffered = ""
                columns[1] = "{}{}{}".format(int_to_b64(position + offset),
                                             index_type,
                                             remaining)
                buffered += FIELDSEPARATOR.join(columns) + "\n"
                bufcounter += 1
            shifted_count += bufcounter
            outfh.write(buffered) #last block
    return shifted_count


def encode_delimiter_re(delimiter):
    """
    encode a delimiter regexp, replacing any ^ and $ with newlines
    """
    if delimiter[0] == "^" or delimiter[-1] == "$":
        delimiter = list(delimiter)
        if delimiter[0] == "^":
            delimiter[0] = "\n"
        if delimiter[-1] == "$":
            delimiter[-1] = "\n"
        delimiter = "".join(delimiter)
    return "{}".format(delimiter).encode('UTF-8')


def postprocess_entry(entry, mode):
    """
    perform any compression, encryption, checksum calculation as needed
    """
    if mode['xsanity']:
        entry['checksum'] = zlib.crc32(entry['full'])
    if mode['encrypt']:
        entry['iv'] = generate_iv() #generate and store new iv
        cipher = init_cipher(mode['key'], entry['iv'])
        if mode['compress']:
            entry['full'] = cipher.encrypt(deflate(entry['full'],
                                                   compresslevel=mode['compresslevel']))
            entry['length'] = len(entry['full']) #probably smaller, need to recompute
        else:
            entry['full'] = cipher.encrypt(entry['full'])
        mode['outffh'].write(entry['full'])
    elif mode['compress']: #compressed but not encrypted
        entry['full'] = deflate(entry['full'], compresslevel=mode['compresslevel'])
        entry['length'] = len(entry['full']) #probably smaller, need to recompute
        mode['outffh'].write(entry['full'])
    entry['length_b64'] = int_to_b64(entry['length'])


def entry_generator(path, delimiter, buffersize=None):
    """
    splits a file according to a delimiter, returning a generator of entries
    terminating in the delimiter
    """
    delimiter = encode_delimiter_re(delimiter)
    splen = len(delimiter)
    buf = b''
    with open(path, 'rb') as fh:
        data = fh.read(buffersize)
        while data:
            #buf += fh.read(buffersize)
            buf += data
            start = end = 0
            while end >= 0:# and len(buf)>=splen:
                start, end = end, buf.find(delimiter, end)
                if end != -1:
                    end += splen
                    yield buf[start:end], end - start
                else:
                    buf = buf[start:]
                    break
            data = fh.read(buffersize)
        #yield buf #if final content after terminator is desired


def compute_split_positions(filename, splitsize, delimiter=None):
    """
    compute and return start positions and sizes of chunks of splitsize
    without breaking lines or without breaking entries (if delimiter provided)
    """
    split_file_sizes = [0] #init with 0 size (this will be removed later)
    input_filesize = os.path.getsize(filename)

    split_file_startpos = [0] #first split start position is 0 (start of file)
    split_file_sizes = list() #init
    if delimiter is None:
        with open(filename, 'r', 1) as inputfh:
            split_file_end = find_next_newline(inputfh, input_filesize,
                                               splitsize)
            split_file_sizes.append(split_file_end)
            while split_file_end != input_filesize:
                split_file_start = split_file_end #of previous block
                split_file_startpos.append(split_file_start)
                split_file_end = find_next_newline(inputfh, input_filesize,
                                                   split_file_start + splitsize)
                split_file_sizes.append(split_file_end - split_file_start)
    else:
        delimiter = encode_delimiter_re(delimiter)
        bufsize = max(1024, min(1024*splitsize//65536, 1024*1024))
        with open(filename, 'rb', 1) as inputfh:
            split_file_end = find_next_delimiter(inputfh, input_filesize,
                                                 splitsize, delimiter, bufsize)
            split_file_sizes.append(split_file_end)
            while split_file_end != input_filesize:
                split_file_start = split_file_end #of previous block
                split_file_startpos.append(split_file_start)
                split_file_end = find_next_delimiter(inputfh, input_filesize,
                                                     split_file_start + splitsize,
                                                     delimiter, bufsize)
                split_file_sizes.append(split_file_end - split_file_start)

    return split_file_startpos, split_file_sizes


def split_file(filename, splitsize, outprefix, delimiter=None):
    """
    split a file into chunks of splitsize without breaking lines
    returns list of chunk filenames and list of the chunk filesizes
    """
    split_files = list()
    split_file_startpos, split_file_sizes = compute_split_positions(filename, splitsize, delimiter)

    split_file_count = len(split_file_sizes)
    suffixlength = len(str(split_file_count))

    with open(filename, 'rb') as inputfh:
        for chunknum in range(split_file_count):
            chunk_suffix = str(chunknum).zfill(suffixlength)
            splitfile = outprefix + chunk_suffix
            split_files.append(splitfile)
            with open(splitfile, 'wb') as outfh:
                inputfh.seek(split_file_startpos[chunknum])
                ##straight, unbuffered, will read whole file in memory:
                #outfh.write(inputfh.read(split_file_sizes[chunknum]))
                ##buffered:
                blockcount, remainder = divmod(split_file_sizes[chunknum], BUFFERSIZE)
                for _ in range(blockcount):
                    buffered = inputfh.read(BUFFERSIZE)
                    outfh.write(buffered)
                outfh.write(inputfh.read(remainder))
    return split_files, split_file_sizes


def buffered_readwrite(inputfile, outputfile, startpos, size):
    """
    extract a piece of inputfile of given size from given position
    and write it to outputfile
    """
    with open(inputfile, 'rb') as inputfh:
        with open(outputfile, 'wb') as outfh:
            inputfh.seek(startpos)
            blockcount, remainder = divmod(size, BUFFERSIZE)
            for _ in range(blockcount):
                buffered = inputfh.read(BUFFERSIZE)
                outfh.write(buffered)
            outfh.write(inputfh.read(remainder))
            outfh.flush()


def check_iofiles(read_filenames, write_filenames):
    """
    check for ability to open input/output filenames
    """
    if read_filenames is not None:
        for filename in read_filenames:
            try:
                inputfh = open(filename, 'r')
                inputfh.close()
            except FileNotFoundError:
                eprint("    => ERROR: Cannot open file '{}' for reading".format(filename))
                sys.exit(2)

    if write_filenames is not None:
        for filename in write_filenames:
            if os.path.isfile(filename):
                eprint("    => ERROR: file '{}' exists".format(filename))
                eprint("       please remove it as we refuse to overwrite it!")
                sys.exit(1)
            try:
                myoutputfh = open(filename, 'w')
                myoutputfh.close()
            except PermissionError:
                eprint("    => ERROR: Cannot open file '{}' for writing".format(filename))
                sys.exit(1)
            #eprint("deleting {}".format(filename)) #DEBUG
            delete_files([filename])


def siprefix2num(numberstring):
    """
    Returns bytes from a metrix suffixed string
    """
    if numberstring == '0':
        return(0)

    prefix = {"k": 1024, "m": 1024**2, "g": 1024**3, "t": 1024**4, "p": 1024**5}
    numberstring = numberstring.lower()
    try:
        return int(numberstring[:-1]) * prefix[numberstring[-1]]
    except KeyError:
        # no or unknown meter-prefix
        return int(numberstring)


def elapsed_time(start_time, work_done=None):
    """
    compute elapsed time from given start_time (in seconds) and return
    well formatted string. If work_done argument given, compute speed of the process
    """
    process_time = time.time() - start_time
    minutes, seconds = divmod(process_time, 60)
    hours, minutes = divmod(minutes, 60)
    if work_done is None:
        return("{:02.0f}h {:02.0f}m {:02.0f}s".format(
            hours, minutes, seconds))
    process_speed = str(round(work_done / process_time, 2))
    return "{:02.0f}h {:02.0f}m {:02.0f}s".format(hours, minutes, seconds), process_speed


def eprint(*myargs, **kwargs):
    """
    print to stderr, useful for error messages and to not clobber stdout
    """
    print(*myargs, file=sys.stderr, **kwargs)
